# gen_write_callsite_file.py

from __future__ import absolute_import

import re
from fortlab.resolver import base_classes, statements, block_statements, typedecl_statements
from fortlab.kgplugin import Kgen_Plugin

from .gencore_utils import STATE_PBLOCK_USE_PART, STATE_PBLOCK_DECL_PART, STATE_PBLOCK_EXEC_PART, \
    STATE_PBLOCK_CONTAINS_PART, STATE_PBLOCK_SUBP_PART, STATE_PBLOCK_WRITE_IN_EXTERNS, STATE_PBLOCK_WRITE_IN_LOCALS, \
    STATE_PBLOCK_WRITE_OUT_EXTERNS, STATE_PBLOCK_WRITE_OUT_LOCALS, STATE_TBLOCK_USE_PART, STATE_TBLOCK_DECL_PART, \
    STATE_TBLOCK_CONTAINS_PART, STATE_TBLOCK_SUBP_PART, STATE_PBLOCK_WRITE_IN_ARGS, state_gencore_contains

BEFORE_CALLSITE = 'before_callsite'
AFTER_CALLSITE = 'after_callsite'

class Gen_S_Callsite_File(Kgen_Plugin):
    def __init__(self):
        self.frame_msg = None

    # registration
    def register(self, msg):
        self.frame_msg = msg
        callsite_stmts = getinfo('callsite_stmts')

        # register initial events
        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.STATE, GENERATION_STAGE.NODE_CREATED, \
            callsite_stmts[0], None, self.create_callsite_parts1)

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.STATE, GENERATION_STAGE.FINISH_PROCESS, \
            callsite_stmts[0], None, self.invalid_kernel_stmts)

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.STATE, GENERATION_STAGE.BEGIN_PROCESS, \
            callsite_stmts[0], self.is_openmp, self.update_omp_directive)

#        if isinstance(callsite_stmts[-1], base_classes.EndStatement):
#            self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.STATE, GENERATION_STAGE.NODE_CREATED, \
#                callsite_stmts[-1].parent, None, self.create_callsite_parts2)
#        else:
#            self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.STATE, GENERATION_STAGE.NODE_CREATED, \
#                callsite_stmts[-1], None, self.create_callsite_parts2)

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.STATE, GENERATION_STAGE.NODE_CREATED, \
            getinfo('parentblock_stmt'), None, self.create_parentblock_parts)

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.STATE, GENERATION_STAGE.NODE_CREATED, \
            getinfo('topblock_stmt'), None, self.create_topblock_parts)


    def is_openmp(self, node):
        return  getinfo('is_openmp_app')

    def invalid_kernel_stmts(self, node):
        kernel_stmts = getinfo('callsite_stmts')

        for stmt in kernel_stmts:
            stmt.genspair.kgen_forced_line = False

    def update_omp_directive(self, node):
        stmt = node.kgen_stmt
        pstmt = None
        if hasattr(stmt, 'parent'):
            pstmt = stmt.parent
        omp_parallel = None
        omp_shared = None

        if pstmt:
            for s in pstmt.content:
                if isinstance(s, statements.Comment):
                    line = s.item.comment
                    if omp_parallel is None:
                        for clause in [ r'do', r'for', r'sections', r'workshare', '' ]:
                            pmatch = re.match(r'^\s*!\$omp\s+parallel\s+%s\b'%clause, line, re.I)
                            if pmatch:
                                omp_parallel = s
                                break
                    if omp_shared is None:
                        smatch = re.match(r'^\s*!\$omp[&]*\s+[\s\w\(\)]*shared\s*\(\b', line, re.I)
                        if smatch:
                            omp_shared = s
                    if omp_parallel and omp_shared:
                        break
                    if s is stmt:
                        break
            if omp_parallel:
                kgen_vars = []
                kgen_vars.append( [ 'kgen_istrue', 'kgen_count', 'kgen_mymid', 'kgen_msize', 'kgen_osize' ] )
                kgen_vars.append( [ 'kgen_ierr', 'kgen_array_sum', 'kgen_realnum' ] )
                kgen_vars.append( [ 'kgen_isstop', 'kgen_filepath', 'kgen_lockpath', 'kgen_islast', 'kgen_issave' ] )
                kgen_vars.append( [ 'kgen_ischecked', 'kgen_invoke', 'kgen_last_invoke' ] )
                new_line = []
                if omp_shared:
                    new_line.append('%s &'%smatch.group()) 
                    for kvars in kgen_vars:
                        new_line.append('!$omp %s , &'%','.join(kvars))
                    pos = omp_shared.item.comment.find(smatch.group())
                    if pos>=0:
                        new_line.append('!$omp %s'%omp_shared.item.comment[len(smatch.group())+pos:])
                    else:
                        raise  Exception('DEBUG: pos error')
                    omp_shared.genspair.kgen_forced_line = '\n'.join(new_line)
                else:
                    new_line.append('%s &'%pmatch.group()) 
                    if len(kgen_vars)==1:
                        new_line.append('!$omp shared ( %s ) &'%','.join(kgen_vars[0]))
                    else:
                        new_line.append('!$omp shared ( %s , &'%','.join(kgen_vars[0]))
                        if len(kgen_vars)==2:
                            new_line.append('!$omp %s ) &'%','.join(kgen_vars[1]))
                        else:
                            for kvars in kgen_vars[1:-1]:
                                new_line.append('!$omp %s , &'%','.join(kvars))
                            new_line.append('!$omp %s ) &'%','.join(kgen_vars[-1]))
                    pos = omp_parallel.item.comment.find(pmatch.group())
                    if pos>=0:
                        remained = omp_parallel.item.comment[len(pmatch.group())+pos:].strip()
                        if len(remained)==0: 
                            new_line[-1] = new_line[-1][:-2]
                        else:
                            new_line.append('!$omp %s'%omp_parallel.item.comment[len(pmatch.group())+pos:])
                    else:
                        raise  Exception('DEBUG: pos error')
                    omp_parallel.genspair.kgen_forced_line = '\n'.join(new_line)
            elif stmt is not getinfo('parentblock_stmt'):
                self.update_omp_directive(pstmt.genspair)

    def create_parentblock_parts(self, node):

        namedpart_link_part(node, STATE_PBLOCK_USE_PART, USE_PART)
        namedpart_link_part(node, STATE_PBLOCK_DECL_PART, DECL_PART)
        namedpart_link_part(node, STATE_PBLOCK_EXEC_PART, EXEC_PART)
        namedpart_link_part(node, STATE_PBLOCK_CONTAINS_PART, CONTAINS_PART)
        namedpart_link_part(node, STATE_PBLOCK_SUBP_PART, SUBP_PART)

        #attrs = {'name':'IEEE_ARITHMETIC', 'nature': 'INTRINSIC', 'isonly': True, 'items':['ieee_is_normal']}
        #namedpart_append_gensnode(node.kgen_kernel_id, STATE_PBLOCK_USE_PART, statements.Use, attrs=attrs)

        # ensure contains
        checks = lambda n: n.kgen_match_class==statements.Contains
        if not node in state_gencore_contains and not part_has_node(node, CONTAINS_PART, checks):
            part_append_comment(node, CONTAINS_PART, '')
            part_append_gensnode(node, CONTAINS_PART, statements.Contains)
            part_append_comment(node, CONTAINS_PART, '')
            state_gencore_contains.append(node)

        #FUNCTION kgen_get_newunit() RESULT(new_unit)
        attrs = {'name': 'kgen_get_newunit', 'result': 'new_unit' }
        unitsubr = part_append_gensnode(node, SUBP_PART, block_statements.Function, attrs=attrs)
        part_append_comment(node, SUBP_PART, '')

        attrs = {'type_spec': 'INTEGER', 'attrspec': ['PARAMETER'], 'entity_decls': ['UNIT_MIN=100', 'UNIT_MAX=1000000']}
        part_append_gensnode(unitsubr, DECL_PART, typedecl_statements.Integer, attrs=attrs)
    
        attrs = {'type_spec': 'LOGICAL', 'entity_decls': ['is_opened']}
        part_append_gensnode(unitsubr, DECL_PART, typedecl_statements.Logical, attrs=attrs)

        attrs = {'type_spec': 'INTEGER', 'entity_decls': ['nunit', 'new_unit', 'counter']}
        part_append_gensnode(unitsubr, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        part_append_comment(unitsubr, DECL_PART, '')

        attrs = {'variable': 'new_unit', 'sign': '=', 'expr': '-1'}
        part_append_gensnode(unitsubr, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'loopcontrol': 'counter=UNIT_MIN, UNIT_MAX'}
        docnt = part_append_gensnode(unitsubr, EXEC_PART, block_statements.Do, attrs=attrs)

        attrs = {'specs': ['UNIT=counter', 'OPENED=is_opened']}
        part_append_gensnode(docnt, EXEC_PART, statements.Inquire, attrs=attrs)

        attrs = {'expr': '.NOT. is_opened'}
        ifopen = part_append_gensnode(docnt, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'variable': 'new_unit', 'sign': '=', 'expr': 'counter'}
        part_append_gensnode(ifopen, EXEC_PART, statements.Assignment, attrs=attrs)

        part_append_gensnode(ifopen, EXEC_PART, statements.Exit, attrs=attrs)

        #SUBROUTINE kgen_init_vars
        attrs = {'name': 'kgen_init_vars', 'args': ['mpi_s', 'mpi_e', 'omp_s', 'omp_e', 'invoke_e', 'msize', \
            'osize', 'lockpath', 'last_invoke', 'isstop']}
        initvarsubr = part_append_gensnode(node, SUBP_PART, block_statements.Subroutine, attrs=attrs)
        part_append_comment(node, SUBP_PART, '')

        attrs = {'type_spec': 'INTEGER', 'attrspec': ['INTENT(IN)'], 'entity_decls': \
            ['mpi_s', 'mpi_e', 'omp_s', 'omp_e', 'invoke_e', 'msize', 'osize'], }
        part_append_gensnode(initvarsubr, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        attrs = {'type_spec': 'INTEGER', 'attrspec': ['INTENT(INOUT)', 'DIMENSION(0:osize-1)'], 'entity_decls': ['last_invoke']}
        part_append_gensnode(initvarsubr, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        attrs = {'type_spec': 'LOGICAL', 'attrspec': ['INTENT(INOUT)', 'DIMENSION(0:msize-1,0:osize-1)'], 'entity_decls': ['isstop']}
        part_append_gensnode(initvarsubr, DECL_PART, typedecl_statements.Logical, attrs=attrs)

        attrs = {'type_spec': 'CHARACTER', 'attrspec': ['INTENT(INOUT)', 'DIMENSION(0:msize-1,0:osize-1)'], \
            'entity_decls': ['lockpath'], 'selector':('1024', None)}
        part_append_gensnode(initvarsubr, DECL_PART, typedecl_statements.Character, attrs=attrs)

        attrs = {'type_spec': 'INTEGER', 'entity_decls': ['mpi_idx', 'openmp_idx', 'temp_unit', 'ierr']}
        part_append_gensnode(initvarsubr, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        attrs = {'loopcontrol': 'mpi_idx=mpi_s,mpi_e'}
        dompi = part_append_gensnode(initvarsubr, EXEC_PART, block_statements.Do, attrs=attrs)

        attrs = {'loopcontrol': 'openmp_idx=omp_s,omp_e'}
        doopenmp = part_append_gensnode(dompi, EXEC_PART, block_statements.Do, attrs=attrs)

        attrs = {'specs': [ 'lockpath(mpi_idx, openmp_idx)', 'FMT="(A,I0,A,I0)"'], \
            'items': ['"%s/done."'%getinfo('kernel_path'), 'mpi_idx', '"."', 'openmp_idx']}
        part_append_gensnode(doopenmp, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'variable': 'isstop(mpi_idx, openmp_idx)', 'sign': '=', 'expr': '.FALSE.'}
        part_append_gensnode(doopenmp, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'variable': 'last_invoke(openmp_idx)', 'sign': '=', 'expr': 'MAX( last_invoke(openmp_idx), invoke_e)'}
        part_append_gensnode(doopenmp, EXEC_PART, statements.Assignment, attrs=attrs)

        #attrs = {'variable': 'temp_unit', 'sign': '=', 'expr': 'kgen_get_newunit()'}
        #part_append_gensnode(doopenmp, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'specs': ['NEWUNIT=temp_unit', 'FILE=lockpath(mpi_idx, openmp_idx)', 'STATUS="OLD"', 'IOSTAT=ierr']}
        part_append_gensnode(doopenmp, EXEC_PART, statements.Open, attrs=attrs)

        attrs = {'expr': 'ierr .EQ. 0'}
        ifopenmp = part_append_gensnode(doopenmp, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'specs': [ 'UNIT=temp_unit', 'STATUS="DELETE"' ]}
        part_append_gensnode(ifopenmp, EXEC_PART, statements.Close, attrs=attrs)

        #SUBROUTINE kgen_check_save
        attrs = {'name': 'kgen_check_save', 'args': ['mpi_s', 'mpi_e', 'omp_s', 'omp_e', 'invoke_s', 'invoke_e', 'mymid', \
            'myoid', 'osize', 'invoke', 'last_invoke', 'issave', 'islast']}
        checksavesubr = part_append_gensnode(node, SUBP_PART, block_statements.Subroutine, attrs=attrs)
        part_append_comment(node, SUBP_PART, '')

        attrs = {'type_spec': 'INTEGER', 'attrspec': ['INTENT(IN)'], 'entity_decls': \
            ['mpi_s', 'mpi_e', 'omp_s', 'omp_e', 'invoke_s', 'invoke_e', 'osize', 'mymid', 'myoid'], }
        part_append_gensnode(checksavesubr, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        attrs = {'type_spec': 'INTEGER', 'attrspec': ['INTENT(IN)', 'DIMENSION(0:osize-1)'], 'entity_decls': ['invoke', 'last_invoke']}
        part_append_gensnode(checksavesubr, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        attrs = {'type_spec': 'LOGICAL', 'attrspec': ['INTENT(OUT)', 'DIMENSION(0:osize-1)'], 'entity_decls': ['issave', 'islast']}
        part_append_gensnode(checksavesubr, DECL_PART, typedecl_statements.Logical, attrs=attrs)

        attrs = {'expr': '(mymid .GE. mpi_s) .AND. (mymid .LE. mpi_e)'}
        ifmpi = part_append_gensnode(checksavesubr, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'expr': '(myoid .GE. omp_s) .AND. (myoid .LE. omp_e)'}
        ifopenmp = part_append_gensnode(ifmpi, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'expr': '(invoke(myoid) .GE. invoke_s) .AND. (invoke(myoid) .LE. invoke_e)'}
        ifinvoke = part_append_gensnode(ifopenmp, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'variable': 'issave(myoid)', 'sign': '=', 'expr': '.TRUE.'}
        part_append_gensnode(ifinvoke, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'expr': 'invoke(myoid) .GE. last_invoke(myoid)'}
        iflast = part_append_gensnode(ifopenmp, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'variable': 'islast(myoid)', 'sign': '=', 'expr': '.TRUE.'}
        part_append_gensnode(iflast, EXEC_PART, statements.Assignment, attrs=attrs)

        #SUBROUTINE kgen_check_stop
        attrs = {'name': 'kgen_check_stop', 'args': ['mpi_s', 'mpi_e', 'omp_s', 'omp_e', 'msize', \
            'osize', 'myoid', 'lockpath', 'isstop']}
        checkstopsubr = part_append_gensnode(node, SUBP_PART, block_statements.Subroutine, attrs=attrs)
        part_append_comment(node, SUBP_PART, '')

        attrs = {'type_spec': 'INTEGER', 'attrspec': ['INTENT(IN)'], 'entity_decls': \
            ['mpi_s', 'mpi_e', 'omp_s', 'omp_e', 'msize', 'osize', 'myoid'] }
        part_append_gensnode(checkstopsubr, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        #attrs = {'type_spec': 'LOGICAL', 'entity_decls': ['isexist', 'isopened'] }
        #part_append_gensnode(checkstopsubr, DECL_PART, typedecl_statements.Logical, attrs=attrs)

        attrs = {'type_spec': 'CHARACTER', 'attrspec': ['INTENT(IN)', 'DIMENSION(0:msize-1,0:osize-1)'], \
            'entity_decls': ['lockpath'], 'selector':('1024', None)}
        part_append_gensnode(checkstopsubr, DECL_PART, typedecl_statements.Character, attrs=attrs)

        attrs = {'type_spec': 'LOGICAL', 'attrspec': ['INTENT(OUT)', 'DIMENSION(0:msize-1,0:osize-1)'], 'entity_decls': ['isstop']}
        part_append_gensnode(checkstopsubr, DECL_PART, typedecl_statements.Logical, attrs=attrs)

        attrs = {'type_spec': 'INTEGER', 'entity_decls': ['mpi_idx', 'openmp_idx', 'ierr', 'myunit']}
        part_append_gensnode(checkstopsubr, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        attrs = {'loopcontrol': 'mpi_idx=mpi_s,mpi_e'}
        dompi = part_append_gensnode(checkstopsubr, EXEC_PART, block_statements.Do, attrs=attrs)

        attrs = {'loopcontrol': 'openmp_idx=omp_s,omp_e'}
        doopenmp = part_append_gensnode(dompi, EXEC_PART, block_statements.Do, attrs=attrs)

        attrs = {'expr': '.NOT. isstop(mpi_idx, openmp_idx)'}
        ifstop = part_append_gensnode(doopenmp, EXEC_PART, block_statements.IfThen, attrs=attrs)

        #attrs = {'variable': 'myunit', 'sign': '=', 'expr': 'kgen_get_newunit()'}
        #part_append_gensnode(ifstop, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'specs': ['NEWUNIT=myunit', 'FILE=TRIM(ADJUSTL(lockpath(mpi_idx, openmp_idx)))', \
            'STATUS="OLD"', 'ACTION="READ"', 'IOSTAT=ierr']}
        part_append_gensnode(ifstop, EXEC_PART, statements.Open, attrs=attrs)

        attrs = {'expr': 'ierr .EQ. 0'}
        ifopen = part_append_gensnode(ifstop, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'variable': 'isstop(mpi_idx, openmp_idx)', 'sign': '=', 'expr': '.TRUE.'}
        part_append_gensnode(ifopen, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'specs': [ 'UNIT=myunit' ]}
        part_append_gensnode(ifopen, EXEC_PART, statements.Close, attrs=attrs)

        #SUBROUTINE kgen_write_list
        attrs = {'name': 'kgen_write_list', 'args': ['myunit', 'mpi_s', 'mpi_e', 'omp_s', 'omp_e', 'invoke_s', 'invoke_e']}
        checkstopsubr = part_append_gensnode(node, SUBP_PART, block_statements.Subroutine, attrs=attrs)
        part_append_comment(node, SUBP_PART, '')

        attrs = {'type_spec': 'INTEGER', 'attrspec': ['INTENT(IN)'], 'entity_decls': \
            ['myunit', 'mpi_s', 'mpi_e', 'omp_s', 'omp_e', 'invoke_s', 'invoke_e'] }
        part_append_gensnode(checkstopsubr, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        attrs = {'type_spec': 'INTEGER', 'entity_decls': ['mpi_idx', 'openmp_idx', 'invoke_idx', 'temp_unit', 'ierr']}
        part_append_gensnode(checkstopsubr, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        attrs = {'type_spec': 'CHARACTER', 'selector': ( '16', None ), 'entity_decls': ['mpi_str', 'openmp_str', 'invoke_str']}
        part_append_gensnode(checkstopsubr, DECL_PART, typedecl_statements.Character, attrs=attrs)

        attrs = {'loopcontrol': 'mpi_idx=mpi_s,mpi_e'}
        dompi = part_append_gensnode(checkstopsubr, EXEC_PART, block_statements.Do, attrs=attrs)

        attrs = {'specs': ['mpi_str', '"(I16)"'], 'items': [ 'mpi_idx' ]}
        part_append_gensnode(dompi, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'loopcontrol': 'openmp_idx=omp_s,omp_e'}
        doopenmp = part_append_gensnode(dompi, EXEC_PART, block_statements.Do, attrs=attrs)

        attrs = {'specs': ['openmp_str', '"(I16)"'], 'items': [ 'openmp_idx' ]}
        part_append_gensnode(doopenmp, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'loopcontrol': 'invoke_idx=invoke_s,invoke_e'}
        doinvoke = part_append_gensnode(doopenmp, EXEC_PART, block_statements.Do, attrs=attrs)

        attrs = {'specs': ['invoke_str', '"(I16)"'], 'items': [ 'invoke_idx' ]}
        part_append_gensnode(doinvoke, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'specs': ['UNIT = myunit', 'FMT="(A)"'], 'items': ['"%s." // TRIM(ADJUSTL(mpi_str)) // "." \
// TRIM(ADJUSTL(openmp_str)) // "." // TRIM(ADJUSTL(invoke_str))'%getinfo('kernel_name')]}
        part_append_gensnode(doinvoke, EXEC_PART, statements.Write, attrs=attrs)

        #attrs = {'variable': 'temp_unit', 'sign': '=', 'expr': 'kgen_get_newunit()'}
        #part_append_gensnode(doopenmp, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'specs': ['NEWUNIT=temp_unit', 'FILE="%s/done." // TRIM(ADJUSTL(mpi_str)) // "." // TRIM(ADJUSTL(openmp_str))'% \
            getinfo('kernel_path'), 'STATUS="OLD"', 'IOSTAT=ierr']}
        part_append_gensnode(doopenmp, EXEC_PART, statements.Open, attrs=attrs)

        attrs = {'expr': 'ierr .EQ. 0'}
        ifopenmp = part_append_gensnode(doopenmp, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'specs': [ 'UNIT=temp_unit', 'STATUS="DELETE"' ]}
        part_append_gensnode(ifopenmp, EXEC_PART, statements.Close, attrs=attrs)


        #SUBROUTINE kgen_error_stop( ierr, errmsg )
        attrs = {'name': 'kgen_error_stop', 'args': ['ierr', 'errmsg']}
        stopsubr = part_append_gensnode(node, SUBP_PART, block_statements.Subroutine, attrs=attrs)
        part_append_comment(node, SUBP_PART, '')

        attrs = {'type_spec': 'INTEGER', 'attrspec': ['INTENT(IN)'], 'entity_decls': ['ierr']}
        part_append_gensnode(stopsubr, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        attrs = {'type_spec': 'CHARACTER', 'attrspec': ['INTENT(IN)'], 'entity_decls': ['errmsg'], 'selector':('*', None)}
        part_append_gensnode(stopsubr, DECL_PART, typedecl_statements.Character, attrs=attrs)

        if getinfo('is_mpi_app'):
            attrs = {'type_spec': 'INTEGER', 'entity_decls': ['kgen_ierr']}
            part_append_gensnode(stopsubr, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        attrs = {'expr': 'ierr /= 0'}
        iferr = part_append_gensnode(stopsubr, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'items': ['errmsg']}
        part_append_gensnode(iferr, EXEC_PART, statements.Write, attrs=attrs)

        if getinfo('is_mpi_app'):
            attrs = {'designator': 'mpi_abort', 'items': [getinfo('mpi_comm'), '0', 'kgen_ierr']}
            part_append_gensnode(iferr, EXEC_PART, statements.Call, attrs=attrs)
        else:
            attrs = {'code': '1'}
            part_append_gensnode(iferr, EXEC_PART, statements.Stop, attrs=attrs)

        #SUBROUTINE kgen_print_counter( counter )
        attrs = {'name': 'kgen_print_counter', 'args': ['counter']}
        cntsubr = part_append_gensnode(node, SUBP_PART, block_statements.Subroutine, attrs=attrs)
        part_append_comment(node, SUBP_PART, '')

        attrs = {'type_spec': 'INTEGER', 'attrspec': ['INTENT(IN)'], 'entity_decls': ['counter']}
        part_append_gensnode(cntsubr, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        attrs = {'items': ['"KGEN writes input state variables at count = "', 'counter']}
        part_append_gensnode(cntsubr, EXEC_PART, statements.Write, attrs=attrs)


    def create_topblock_parts(self, node):
        node.kgen_stmt.top.used4genstate = True

        namedpart_link_part(node, STATE_TBLOCK_USE_PART, USE_PART)
        namedpart_link_part(node, STATE_TBLOCK_DECL_PART, DECL_PART)
        namedpart_link_part(node, STATE_TBLOCK_CONTAINS_PART, CONTAINS_PART)
        namedpart_link_part(node, STATE_TBLOCK_SUBP_PART, SUBP_PART)

        attrs = {'name': 'KGEN'}
        cblock = part_append_gensnode(node.kgen_parent, UNIT_PART, block_statements.BlockData, attrs=attrs)

        attrs = {'type_spec': 'INTEGER', 'entity_decls': ['kgen_mpirank = 0']}
        part_append_gensnode(cblock, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        if getinfo('is_openmp_app'):
            attrs = {'type_spec': 'INTEGER', 'attrspec': [ 'DIMENSION(0:1023)' ], 'entity_decls': ['kgen_openmp_issave = -1']}
            part_append_gensnode(cblock, DECL_PART, typedecl_statements.Integer, attrs=attrs)
        else:
            attrs = {'type_spec': 'INTEGER', 'entity_decls': ['kgen_openmp_issave = -1']}
            part_append_gensnode(cblock, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        attrs = {'items': [ ( 'kgen_state', ('kgen_mpirank', 'kgen_openmp_issave') ) ]}
        part_append_gensnode(cblock, DECL_PART, statements.Common, attrs=attrs)

    def create_callsite_parts1(self, node):

        # declaration part

        if getinfo('is_mpi_app'):
            mpi_use_stmts = getinfo('mpi_use')
            if mpi_use_stmts and len(mpi_use_stmts)>0:
                for mod_name, use_names in mpi_use_stmts:
                    attrs = {'name':mod_name, 'isonly': True, 'items':use_names}
                    namedpart_insert_gensnode(node.kgen_kernel_id, STATE_PBLOCK_USE_PART, statements.Use, 0, attrs=attrs)


        attrs = {'type_spec': 'INTEGER', 'attrspec': [ 'SAVE', 'ALLOCATABLE', 'DIMENSION(:)' ], \
            'entity_decls': ['kgen_invoke', 'kgen_last_invoke']}
        namedpart_insert_gensnode(node.kgen_kernel_id, STATE_PBLOCK_DECL_PART, typedecl_statements.Integer, 0, attrs=attrs)

        attrs = {'type_spec': 'LOGICAL', 'attrspec': [ 'SAVE', 'ALLOCATABLE', 'DIMENSION(:)' ], \
            'entity_decls': ['kgen_islast', 'kgen_issave', 'kgen_ischecked']}
        namedpart_insert_gensnode(node.kgen_kernel_id, STATE_PBLOCK_DECL_PART, typedecl_statements.Logical, 0, attrs=attrs)

        attrs = {'type_spec': 'CHARACTER', 'attrspec': [ 'SAVE', 'ALLOCATABLE', 'DIMENSION(:,:)' ], \
            'entity_decls': ['kgen_filepath', 'kgen_lockpath'], 'selector':('1024', None)}
        namedpart_insert_gensnode(node.kgen_kernel_id, STATE_PBLOCK_DECL_PART, typedecl_statements.Character, 0, attrs=attrs)

        attrs = {'type_spec': 'LOGICAL', 'attrspec': [ 'SAVE', 'ALLOCATABLE', 'DIMENSION(:,:)' ], 'entity_decls': ['kgen_isstop']}
        namedpart_insert_gensnode(node.kgen_kernel_id, STATE_PBLOCK_DECL_PART, typedecl_statements.Logical, 0, attrs=attrs)

        attrs = {'type_spec': 'REAL', 'attrspec': [ 'SAVE' ], 'entity_decls': ['kgen_array_sum', 'kgen_realnum'], 'selector': (None, '8')}
        namedpart_insert_gensnode(node.kgen_kernel_id, STATE_PBLOCK_DECL_PART, typedecl_statements.Real, 0, attrs=attrs)

        attrs = {'type_spec': 'INTEGER', 'attrspec': [ 'SAVE' ], 'entity_decls': ['kgen_mymid', 'kgen_msize', 'kgen_osize', 'kgen_ierr']}
        namedpart_insert_gensnode(node.kgen_kernel_id, STATE_PBLOCK_DECL_PART, typedecl_statements.Integer, 0, attrs=attrs)

        if getinfo('is_openmp_app'):
            attrs = {'type_spec': 'INTEGER', 'attrspec': [ 'SAVE', 'ALLOCATABLE', 'DIMENSION(:)' ], 'entity_decls': ['kgen_save_unit', 'kgen_stopunit']}
            namedpart_insert_gensnode(node.kgen_kernel_id, STATE_PBLOCK_DECL_PART, typedecl_statements.Integer, 0, attrs=attrs)

            attrs = {'type_spec': 'INTEGER', 'entity_decls': ['OMP_GET_NUM_THREADS', 'OMP_GET_THREAD_NUM']}
            namedpart_insert_gensnode(node.kgen_kernel_id, STATE_PBLOCK_DECL_PART, typedecl_statements.Integer, 0, attrs=attrs)

            attrs = {'type_spec': 'INTEGER', 'entity_decls': ['kgen_unit']}
            namedpart_insert_gensnode(node.kgen_kernel_id, STATE_PBLOCK_DECL_PART, typedecl_statements.Integer, 0, attrs=attrs)

            #attrs = {'type_spec': 'LOGICAL', 'entity_decls': ['OMP_IN_PARALLEL']}
            #namedpart_insert_gensnode(node.kgen_kernel_id, STATE_PBLOCK_DECL_PART, typedecl_statements.Logical, 0, attrs=attrs)
        else:
            attrs = {'type_spec': 'INTEGER', 'attrspec': [ 'SAVE' ], 'entity_decls': ['kgen_unit', 'kgen_stopunit']}
            namedpart_insert_gensnode(node.kgen_kernel_id, STATE_PBLOCK_DECL_PART, typedecl_statements.Integer, 0, attrs=attrs)

        attrs = {'type_spec': 'INTEGER', 'entity_decls': ['kgen_count']}
        namedpart_insert_gensnode(node.kgen_kernel_id, STATE_PBLOCK_DECL_PART, typedecl_statements.Integer, 0, attrs=attrs)

        attrs = {'type_spec': 'LOGICAL', 'entity_decls': ['kgen_istrue']}
        namedpart_insert_gensnode(node.kgen_kernel_id, STATE_PBLOCK_DECL_PART, typedecl_statements.Logical, 0, attrs=attrs)

        attrs = {'items': [ ( 'kgen_state', ('kgen_mpirank', 'kgen_openmp_issave') ) ]}
        namedpart_insert_gensnode(node.kgen_kernel_id, STATE_PBLOCK_DECL_PART, statements.Common, 0, attrs=attrs)

        attrs = {'type_spec': 'INTEGER', 'entity_decls': ['kgen_mpirank']}
        namedpart_insert_gensnode(node.kgen_kernel_id, STATE_PBLOCK_DECL_PART, typedecl_statements.Integer, 0, attrs=attrs)

        if getinfo('is_openmp_app'):
            attrs = {'type_spec': 'INTEGER', 'attrspec': [ 'DIMENSION(0:1023)' ], 'entity_decls': ['kgen_openmp_issave']}
            namedpart_insert_gensnode(node.kgen_kernel_id, STATE_PBLOCK_DECL_PART, typedecl_statements.Integer, 0, attrs=attrs)
        else:
            attrs = {'type_spec': 'INTEGER', 'entity_decls': ['kgen_openmp_issave']}
            namedpart_insert_gensnode(node.kgen_kernel_id, STATE_PBLOCK_DECL_PART, typedecl_statements.Integer, 0, attrs=attrs)

        namedpart_insert_comment(node.kgen_kernel_id, STATE_PBLOCK_DECL_PART, 0, 'kgen variables')

        index, partname, part = get_part_index(node)

        namedpart_create_subpart(node.kgen_parent, BEFORE_CALLSITE, EXEC_PART, index=index)

        # initialize variables
        namedpart_append_comment(node.kgen_kernel_id, BEFORE_CALLSITE, 'START OF KGEN REGION')

        if getinfo('is_openmp_app'):

            #attrs = {'expr': 'OMP_IN_PARALLEL()'}
            #ifomp = namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, block_statements.IfThen, attrs=attrs)
            #part_append_comment(ifomp, EXEC_PART, 'CRITICAL (kgen_init)', style='openmp')
            namedpart_append_comment(node.kgen_kernel_id, BEFORE_CALLSITE, 'CRITICAL (kgen_init)', style='openmp')

        attrs = {'expr': '.NOT. ALLOCATED(kgen_isstop)'}
        ifinit = namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, block_statements.IfThen, attrs=attrs)

        if getinfo('is_openmp_app'):
            attrs = {'variable': 'kgen_osize', 'sign': '=', 'expr': 'OMP_GET_NUM_THREADS()'}
            part_append_gensnode(ifinit, EXEC_PART, statements.Assignment, attrs=attrs)

            attrs = {'items': ['kgen_save_unit(0:kgen_osize-1)']}
            part_append_gensnode(ifinit, EXEC_PART, statements.Allocate, attrs=attrs)

            attrs = {'items': ['kgen_stopunit(0:kgen_osize-1)']}
            part_append_gensnode(ifinit, EXEC_PART, statements.Allocate, attrs=attrs)

            attrs = {'variable': 'kgen_save_unit(:)', 'sign': '=', 'expr': '-1'}
            part_append_gensnode(ifinit, EXEC_PART, statements.Assignment, attrs=attrs)

            attrs = {'variable': 'kgen_stopunit(:)', 'sign': '=', 'expr': '-1'}
            part_append_gensnode(ifinit, EXEC_PART, statements.Assignment, attrs=attrs)

        else:
            attrs = {'variable': 'kgen_osize', 'sign': '=', 'expr': '1'}
            part_append_gensnode(ifinit, EXEC_PART, statements.Assignment, attrs=attrs)

            attrs = {'variable': 'kgen_unit', 'sign': '=', 'expr': '-1'}
            part_append_gensnode(ifinit, EXEC_PART, statements.Assignment, attrs=attrs)

            attrs = {'variable': 'kgen_stopunit', 'sign': '=', 'expr': '-1'}
            part_append_gensnode(ifinit, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'items': ['kgen_ischecked(0:kgen_osize-1)']}
        part_append_gensnode(ifinit, EXEC_PART, statements.Allocate, attrs=attrs)

        attrs = {'items': ['kgen_islast(0:kgen_osize-1)']}
        part_append_gensnode(ifinit, EXEC_PART, statements.Allocate, attrs=attrs)

        attrs = {'items': ['kgen_issave(0:kgen_osize-1)']}
        part_append_gensnode(ifinit, EXEC_PART, statements.Allocate, attrs=attrs)

        attrs = {'items': ['kgen_invoke(0:kgen_osize-1)']}
        part_append_gensnode(ifinit, EXEC_PART, statements.Allocate, attrs=attrs)

        attrs = {'items': ['kgen_last_invoke(0:kgen_osize-1)']}
        part_append_gensnode(ifinit, EXEC_PART, statements.Allocate, attrs=attrs)


        attrs = {'variable': 'kgen_ischecked(:)', 'sign': '=', 'expr': '.FALSE.'}
        part_append_gensnode(ifinit, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'variable': 'kgen_islast(:)', 'sign': '=', 'expr': '.FALSE.'}
        part_append_gensnode(ifinit, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'variable': 'kgen_issave(:)', 'sign': '=', 'expr': '.FALSE.'}
        part_append_gensnode(ifinit, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'variable': 'kgen_invoke(:)', 'sign': '=', 'expr': '1'}
        part_append_gensnode(ifinit, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'variable': 'kgen_last_invoke(:)', 'sign': '=', 'expr': '1'}
        part_append_gensnode(ifinit, EXEC_PART, statements.Assignment, attrs=attrs)

        if getinfo('is_mpi_app'):
            attrs = {'designator': 'mpi_comm_rank', 'items': [getinfo('mpi_comm'), 'kgen_mymid', 'kgen_ierr']}
            part_append_gensnode(ifinit, EXEC_PART, statements.Call, attrs=attrs)

            attrs = {'designator': 'kgen_error_stop', 'items': ['kgen_ierr', '"mpi_comm_rank is failed"']}
            part_append_gensnode(ifinit, EXEC_PART, statements.Call, attrs=attrs)
            
            attrs = {'designator': 'mpi_comm_size', 'items': [getinfo('mpi_comm'), 'kgen_msize', 'kgen_ierr']}
            part_append_gensnode(ifinit, EXEC_PART, statements.Call, attrs=attrs)

            attrs = {'designator': 'kgen_error_stop', 'items': ['kgen_ierr', '"mpi_comm_size is failed"']}
            part_append_gensnode(ifinit, EXEC_PART, statements.Call, attrs=attrs)

            attrs = {'variable': 'kgen_mpirank', 'sign': '=', 'expr': 'kgen_mymid'}
            part_append_gensnode(ifinit, EXEC_PART, statements.Assignment, attrs=attrs)

        else:
            attrs = {'variable': 'kgen_mymid', 'sign': '=', 'expr': '0'}
            part_append_gensnode(ifinit, EXEC_PART, statements.Assignment, attrs=attrs)

            attrs = {'variable': 'kgen_msize', 'sign': '=', 'expr': '1'}
            part_append_gensnode(ifinit, EXEC_PART, statements.Assignment, attrs=attrs)

            attrs = {'variable': 'kgen_mpirank', 'sign': '=', 'expr': '0'}
            part_append_gensnode(ifinit, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'items': ['kgen_filepath(0:kgen_msize-1, 0:kgen_osize-1)']}
        part_append_gensnode(ifinit, EXEC_PART, statements.Allocate, attrs=attrs)

        attrs = {'items': ['kgen_lockpath(0:kgen_msize-1, 0:kgen_osize-1)']}
        part_append_gensnode(ifinit, EXEC_PART, statements.Allocate, attrs=attrs)

        attrs = {'items': ['kgen_isstop(0:kgen_msize-1, 0:kgen_osize-1)']}
        part_append_gensnode(ifinit, EXEC_PART, statements.Allocate, attrs=attrs)

        attrs = {'variable': 'kgen_isstop(:,:)', 'sign': '=', 'expr': '.TRUE.'}
        part_append_gensnode(ifinit, EXEC_PART, statements.Assignment, attrs=attrs)

        for (mpi_s, mpi_e), (openmp_s, openmp_e), (invoke_s, invoke_e) in getinfo('invocations'):
            attrs = {'designator': 'kgen_init_vars', 'items': ['INT(%s)'%mpi_s.replace('e', '(kgen_msize-1)'), \
                'INT(%s)'%mpi_e.replace('e', '(kgen_msize-1)'), 'INT(%s)'%openmp_s.replace('e', '(kgen_osize-1)'), \
                'INT(%s)'%openmp_e.replace('e', '(kgen_osize-1)'), 'INT(%s)'%invoke_e, 'kgen_msize', 'kgen_osize', 'kgen_lockpath', \
                'kgen_last_invoke', 'kgen_isstop']}
            part_append_gensnode(ifinit, EXEC_PART, statements.Call, attrs=attrs)

        # check save
        if getinfo('is_openmp_app'):

            #attrs = {'expr': 'OMP_IN_PARALLEL()'}
            #ifomp = namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, block_statements.IfThen, attrs=attrs)
            #part_append_comment(ifomp, EXEC_PART, 'END CRITICAL (kgen_init)', style='openmp')
            namedpart_append_comment(node.kgen_kernel_id, BEFORE_CALLSITE, 'END CRITICAL (kgen_init)', style='openmp')

            attrs = {'variable': 'kgen_issave(OMP_GET_THREAD_NUM())', 'sign': '=', 'expr': '.FALSE.'}
            namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, statements.Assignment, attrs=attrs)

            #attrs = {'variable': 'kgen_islast(OMP_GET_THREAD_NUM())', 'sign': '=', 'expr': '.FALSE.'}
            #namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, statements.Assignment, attrs=attrs)

            for (mpi_s, mpi_e), (openmp_s, openmp_e), (invoke_s, invoke_e) in getinfo('invocations'):
                attrs = {'designator': 'kgen_check_save', 'items': ['INT(%s)'%mpi_s.replace('e', '(kgen_msize-1)'), \
                    'INT(%s)'%mpi_e.replace('e', '(kgen_msize-1)'), 'INT(%s)'%openmp_s.replace('e', '(kgen_osize-1)'), \
                    'INT(%s)'%openmp_e.replace('e', '(kgen_osize-1)'), 'INT(%s)'%invoke_s, 'INT(%s)'%invoke_e, \
                    'kgen_mymid', 'OMP_GET_THREAD_NUM()', 'kgen_osize', 'kgen_invoke', 'kgen_last_invoke', \
                    'kgen_issave', 'kgen_islast']}
                namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, statements.Call, attrs=attrs)

            #attrs = {'expr': 'OMP_IN_PARALLEL()'}
            #ifomp = namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, block_statements.IfThen, attrs=attrs)
            #part_append_comment(ifomp, EXEC_PART, 'CRITICAL (kgen_kernel)', style='openmp')


            for gentype, condexpr in getinfo('data')['condition']:
                if gentype == 'set':
                    attrs = {'expr': condexpr}
                elif gentype == 'and':
                    attrs = {'expr': '%s .AND. kgen_issave(OMP_GET_THREAD_NUM())'%condexpr}
                elif gentype == 'or':
                    attrs = {'expr': '%s .OR. kgen_issave(OMP_GET_THREAD_NUM())'%condexpr}
                ifcond = namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, block_statements.IfThen, attrs=attrs)

                attrs = {'variable': 'kgen_issave(OMP_GET_THREAD_NUM())', 'sign': '=', 'expr': '.TRUE.'}
                part_append_gensnode(ifcond, EXEC_PART, statements.Assignment, attrs=attrs)

                part_append_gensnode(ifcond, EXEC_PART, statements.Else, attrs=attrs)

                attrs = {'variable': 'kgen_issave(OMP_GET_THREAD_NUM())', 'sign': '=', 'expr': '.FALSE.'}
                part_append_gensnode(ifcond, EXEC_PART, statements.Assignment, attrs=attrs)

            namedpart_append_comment(node.kgen_kernel_id, BEFORE_CALLSITE, 'CRITICAL (kgen_kernel)', style='openmp')


            attrs = {'expr': 'kgen_issave(OMP_GET_THREAD_NUM())'}
            ifsave = namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, block_statements.IfThen, attrs=attrs)

            l = [ 'kgen_mymid', '"."', 'OMP_GET_THREAD_NUM()', '"."', 'kgen_invoke(OMP_GET_THREAD_NUM())']
            attrs = {'specs': ['kgen_filepath(kgen_mymid, OMP_GET_THREAD_NUM())', 'FMT="(A,I0,A,I0,A,I0)"' ], \
                'items': [ '"%s/%s."'%(getinfo('kernel_path'), getinfo('kernel_name')) ] + l}
            part_append_gensnode(ifsave, EXEC_PART, statements.Write, attrs=attrs)

            #attrs = {'variable': 'kgen_unit', 'sign': '=', 'expr': 'kgen_get_newunit()'}
            #part_append_gensnode(ifsave, EXEC_PART, statements.Assignment, attrs=attrs)

            attrs = {'specs': ['NEWUNIT=kgen_save_unit(OMP_GET_THREAD_NUM())', 'FILE=TRIM(ADJUSTL(kgen_filepath(kgen_mymid, OMP_GET_THREAD_NUM())))', \
                'STATUS="REPLACE"', 'ACCESS="STREAM"', 'FORM="UNFORMATTED"', 'ACTION="WRITE"', 'CONVERT="BIG_ENDIAN"', 'IOSTAT=kgen_ierr']}
            part_append_gensnode(ifsave, EXEC_PART, statements.Open, attrs=attrs)

            attrs = {'variable': 'kgen_unit', 'sign': '=', 'expr': 'kgen_save_unit(OMP_GET_THREAD_NUM())'}
            part_append_gensnode(ifsave, EXEC_PART, statements.Assignment, attrs=attrs)

            attrs = {'designator': 'kgen_error_stop', 'items': ['kgen_ierr', '"File open error: " // TRIM(ADJUSTL(kgen_filepath(kgen_mymid, OMP_GET_THREAD_NUM())))']}
            part_append_gensnode(ifsave, EXEC_PART, statements.Call, attrs=attrs)

            attrs = {'expr': 'kgen_issave(OMP_GET_THREAD_NUM())'}
            ifsave2 = namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, block_statements.IfThen, attrs=attrs)

            attrs = {'variable': 'kgen_openmp_issave(OMP_GET_THREAD_NUM())', 'sign': '=', 'expr': 'kgen_invoke(OMP_GET_THREAD_NUM())'}
            part_append_gensnode(ifsave2, EXEC_PART, statements.Assignment, attrs=attrs)

            part_append_gensnode(ifsave2, EXEC_PART, statements.Else, attrs=attrs)

            attrs = {'variable': 'kgen_openmp_issave(OMP_GET_THREAD_NUM())', 'sign': '=', 'expr': '-1'}
            part_append_gensnode(ifsave2, EXEC_PART, statements.Assignment, attrs=attrs)

            if not getinfo('is_openmp_critical'):
                namedpart_append_comment(node.kgen_kernel_id, BEFORE_CALLSITE, 'END CRITICAL (kgen_kernel)', style='openmp')

        else:
            attrs = {'variable': 'kgen_issave(0)', 'sign': '=', 'expr': '.FALSE.'}
            namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, statements.Assignment, attrs=attrs)

            attrs = {'variable': 'kgen_islast(0)', 'sign': '=', 'expr': '.FALSE.'}
            namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, statements.Assignment, attrs=attrs)

            for (mpi_s, mpi_e), (openmp_s, openmp_e), (invoke_s, invoke_e) in getinfo('invocations'):
                attrs = {'designator': 'kgen_check_save', 'items': ['INT(%s)'%mpi_s.replace('e', '(kgen_msize-1)'), \
                    'INT(%s)'%mpi_e.replace('e', '(kgen_msize-1)'), 'INT(%s)'%openmp_s.replace('e', '(kgen_osize-1)'), \
                    'INT(%s)'%openmp_e.replace('e', '(kgen_osize-1)'), 'INT(%s)'%invoke_s, 'INT(%s)'%invoke_e, \
                    'kgen_mymid', '0', 'kgen_osize', 'kgen_invoke', 'kgen_last_invoke', \
                    'kgen_issave', 'kgen_islast']}
                namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, statements.Call, attrs=attrs)

            for gentype, condexpr in getinfo('data')['condition']:
                if gentype == 'set':
                    attrs = {'expr': condexpr}
                elif gentype == 'and':
                    attrs = {'expr': '%s .AND. kgen_issave(0)'%condexpr}
                elif gentype == 'or':
                    attrs = {'expr': '%s .OR. kgen_issave(0)'%condexpr}
                ifcond = namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, block_statements.IfThen, attrs=attrs)

                attrs = {'variable': 'kgen_issave(0)', 'sign': '=', 'expr': '.TRUE.'}
                part_append_gensnode(ifcond, EXEC_PART, statements.Assignment, attrs=attrs)

                part_append_gensnode(ifcond, EXEC_PART, statements.Else, attrs=attrs)

                attrs = {'variable': 'kgen_issave(0)', 'sign': '=', 'expr': '.FALSE.'}
                part_append_gensnode(ifcond, EXEC_PART, statements.Assignment, attrs=attrs)

            attrs = {'expr': 'kgen_issave(0)'}
            ifsave = namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, block_statements.IfThen, attrs=attrs)

            l = [ 'kgen_mymid', '"."', '0', '"."', 'kgen_invoke(0)']
            attrs = {'specs': ['kgen_filepath(kgen_mymid, 0)', 'FMT="(A,I0,A,I0,A,I0)"' ], \
                'items': [ '"%s/%s."'%(getinfo('kernel_path'), getinfo('kernel_name')) ] + l}
            part_append_gensnode(ifsave, EXEC_PART, statements.Write, attrs=attrs)

            #attrs = {'variable': 'kgen_unit', 'sign': '=', 'expr': 'kgen_get_newunit()'}
            #part_append_gensnode(ifsave, EXEC_PART, statements.Assignment, attrs=attrs)

            attrs = {'specs': ['NEWUNIT=kgen_unit', 'FILE=TRIM(ADJUSTL(kgen_filepath(kgen_mymid, 0)))', \
                'STATUS="REPLACE"', 'ACCESS="STREAM"', 'FORM="UNFORMATTED"', 'ACTION="WRITE"', 'CONVERT="BIG_ENDIAN"', 'IOSTAT=kgen_ierr']}
            part_append_gensnode(ifsave, EXEC_PART, statements.Open, attrs=attrs)

            attrs = {'designator': 'kgen_error_stop', 'items': ['kgen_ierr', '"File open error: " // TRIM(ADJUSTL(kgen_filepath(kgen_mymid, 0)))']}
            part_append_gensnode(ifsave, EXEC_PART, statements.Call, attrs=attrs)

            attrs = {'expr': 'kgen_issave(0)'}
            ifsave2 = namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, block_statements.IfThen, attrs=attrs)

            attrs = {'variable': 'kgen_openmp_issave', 'sign': '=', 'expr': 'kgen_invoke(0)'}
            part_append_gensnode(ifsave2, EXEC_PART, statements.Assignment, attrs=attrs)

            part_append_gensnode(ifsave2, EXEC_PART, statements.Else, attrs=attrs)

            attrs = {'variable': 'kgen_openmp_issave', 'sign': '=', 'expr': '-1'}
            part_append_gensnode(ifsave2, EXEC_PART, statements.Assignment, attrs=attrs)

        namedpart_create_subpart(ifsave, STATE_PBLOCK_WRITE_IN_ARGS, EXEC_PART)
        namedpart_append_comment(node.kgen_kernel_id, STATE_PBLOCK_WRITE_IN_ARGS, '')
        namedpart_append_comment(node.kgen_kernel_id, STATE_PBLOCK_WRITE_IN_ARGS, 'argument input variables')

        namedpart_create_subpart(ifsave, STATE_PBLOCK_WRITE_IN_EXTERNS, EXEC_PART)
        namedpart_append_comment(node.kgen_kernel_id, STATE_PBLOCK_WRITE_IN_EXTERNS, '')
        namedpart_append_comment(node.kgen_kernel_id, STATE_PBLOCK_WRITE_IN_EXTERNS, 'extern input variables')

        namedpart_create_subpart(ifsave, STATE_PBLOCK_WRITE_IN_LOCALS, EXEC_PART)
        namedpart_append_comment(node.kgen_kernel_id, STATE_PBLOCK_WRITE_IN_LOCALS, '')
        namedpart_append_comment(node.kgen_kernel_id, STATE_PBLOCK_WRITE_IN_LOCALS, 'local input variables')

        callsite_stmts = getinfo('callsite_stmts')

        start = callsite_stmts[0].item.span[0]-1
        end = callsite_stmts[-1].item.span[1]
        lines = []
        for line in callsite_stmts[0].top.prep[start:end]:
            match = re.match(r'^\s*!\$omp\s+barrier\b', line, re.I)
            if match:
                lines.append('!%s ! Removed by KGen'%line)
            else:
                lines.append(line)

        lines_str = '\n'.join(lines)
        #dummy_node = part_append_gensnode(ifsave, EXEC_PART, statements.Call)
        dummy_node = namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, statements.Call)
        dummy_node.kgen_stmt = getinfo('dummy_stmt')
        dummy_node.kgen_forced_line = lines_str


        if getinfo('is_openmp_app'):
            if not getinfo('is_openmp_critical'):
                namedpart_append_comment(node.kgen_kernel_id, BEFORE_CALLSITE, 'CRITICAL (kgen_kernel)', style='openmp')

            attrs = {'variable': 'kgen_openmp_issave(OMP_GET_THREAD_NUM())', 'sign': '=', 'expr': '-1'}
            namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, statements.Assignment, attrs=attrs)

            attrs = {'expr': 'kgen_issave(OMP_GET_THREAD_NUM())'}
            ifsave2 = namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, block_statements.IfThen, attrs=attrs)

            attrs = {'variable': 'kgen_unit', 'sign': '=', 'expr': 'kgen_save_unit(OMP_GET_THREAD_NUM())'}
            part_append_gensnode(ifsave2, EXEC_PART, statements.Assignment, attrs=attrs)

        else:
            attrs = {'variable': 'kgen_openmp_issave', 'sign': '=', 'expr': '-1'}
            namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, statements.Assignment, attrs=attrs)

            attrs = {'expr': 'kgen_issave(0)'}
            ifsave2 = namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, block_statements.IfThen, attrs=attrs)

        #import pdb; pdb.set_trace()
        #for stmt in getinfo('callsite_stmts'):
        #    part_append_node(ifsave, EXEC_PART, stmt.genspair)
        
        namedpart_create_subpart(ifsave2, STATE_PBLOCK_WRITE_OUT_EXTERNS, EXEC_PART)
        namedpart_append_comment(node.kgen_kernel_id, STATE_PBLOCK_WRITE_OUT_EXTERNS, '')
        namedpart_append_comment(node.kgen_kernel_id, STATE_PBLOCK_WRITE_OUT_EXTERNS, 'extern output variables')

        namedpart_create_subpart(ifsave2, STATE_PBLOCK_WRITE_OUT_LOCALS, EXEC_PART)
        namedpart_append_comment(node.kgen_kernel_id, STATE_PBLOCK_WRITE_OUT_LOCALS, '')
        namedpart_append_comment(node.kgen_kernel_id, STATE_PBLOCK_WRITE_OUT_LOCALS, 'local output variables')

        if getinfo('is_openmp_app'):
            #attrs = {'specs': [ 'UNIT=kgen_unit(OMP_GET_THREAD_NUM())' ]}
            attrs = {'specs': [ 'kgen_unit' ]}
            part_append_gensnode(ifsave2, EXEC_PART, statements.Close, attrs=attrs)

            attrs = {'items': ['"Collected Kernel Input/Ouput state from: "', 'kgen_mymid', 'OMP_GET_THREAD_NUM()', 'kgen_invoke(OMP_GET_THREAD_NUM())']}
            part_append_gensnode(ifsave2, EXEC_PART, statements.Write, attrs=attrs)

            #attrs = {'expr': 'OMP_IN_PARALLEL()'}
            #ifomp = namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, block_statements.IfThen, attrs=attrs)
            #part_append_comment(ifomp, EXEC_PART, 'END CRITICAL (kgen_kernel)', style='openmp')
            namedpart_append_comment(node.kgen_kernel_id, BEFORE_CALLSITE, 'END CRITICAL (kgen_kernel)', style='openmp')
        else:
            attrs = {'specs': [ 'UNIT=kgen_unit' ]}
            part_append_gensnode(ifsave2, EXEC_PART, statements.Close, attrs=attrs)

            attrs = {'items': ['"Collected Kernel Input/Ouput state from: "', 'kgen_mymid', '0', 'kgen_invoke(0)']}
            part_append_gensnode(ifsave2, EXEC_PART, statements.Write, attrs=attrs)

#
#        part_append_gensnode(ifsave, EXEC_PART, statements.Else) 
#
#        if getinfo('is_openmp_app'):
#            part_append_comment(ifsave, EXEC_PART, 'CRITICAL (kgen_kernel)', style='openmp')
#
#        callsite_stmts = getinfo('callsite_stmts')
#
#        start = callsite_stmts[0].item.span[0]-1
#        end = callsite_stmts[-1].item.span[1]
#        lines = callsite_stmts[0].top.prep[start:end]
#        lines_str = '\n'.join(lines)
#        dummy_node = part_append_gensnode(ifsave, EXEC_PART, statements.Call)
#        dummy_node.kgen_stmt = getinfo('dummy_stmt')
#        dummy_node.kgen_forced_line = lines_str

#        if getinfo('is_openmp_app'):
#            part_append_comment(ifsave, EXEC_PART, 'END CRITICAL (kgen_kernel)', style='openmp')

        # check stop
        if getinfo('is_openmp_app'):

            #attrs = {'expr': 'OMP_IN_PARALLEL()'}
            #ifomp = namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, block_statements.IfThen, attrs=attrs)
            #part_append_comment(ifomp, EXEC_PART, 'CRITICAL (kgen_fini)', style='openmp')
            namedpart_append_comment(node.kgen_kernel_id, BEFORE_CALLSITE, 'CRITICAL (kgen_fini)', style='openmp')

            #attrs = {'items': ['"DEBUG 1: "', 'kgen_ischecked', 'kgen_islast']}
            #namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, statements.Write, attrs=attrs)

            attrs = {'expr': '.NOT. kgen_ischecked(OMP_GET_THREAD_NUM()) .AND. kgen_islast(OMP_GET_THREAD_NUM())'}
            ifstop = namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, block_statements.IfThen, attrs=attrs)

            attrs = {'variable': 'kgen_ischecked(OMP_GET_THREAD_NUM())', 'sign': '=', 'expr': '.TRUE.'}
            part_append_gensnode(ifstop, EXEC_PART, statements.Assignment, attrs=attrs)


            #attrs = {'variable': 'kgen_stopunit', 'sign': '=', 'expr': 'kgen_get_newunit()'}
            #part_append_gensnode(ifstop, EXEC_PART, statements.Assignment, attrs=attrs)

            attrs = {'specs': ['NEWUNIT=kgen_stopunit(OMP_GET_THREAD_NUM())', 'FILE=TRIM(ADJUSTL(kgen_lockpath(kgen_mymid, OMP_GET_THREAD_NUM())))', \
                'STATUS="NEW"', 'IOSTAT=kgen_ierr']}
            part_append_gensnode(ifstop, EXEC_PART, statements.Open, attrs=attrs)

            attrs = {'expr': 'kgen_ierr == 0'}
            ifopen = part_append_gensnode(ifstop, EXEC_PART, block_statements.IfThen, attrs=attrs)

            attrs = {'specs': [ 'UNIT=kgen_stopunit(OMP_GET_THREAD_NUM())', 'STATUS="KEEP"' ]}
            part_append_gensnode(ifopen, EXEC_PART, statements.Close, attrs=attrs)
           
            for (mpi_s, mpi_e), (openmp_s, openmp_e), (invoke_s, invoke_e) in getinfo('invocations'):
                attrs = {'designator': 'kgen_check_stop', 'items': ['INT(%s)'%mpi_s.replace('e', '(kgen_msize-1)'), \
                    'INT(%s)'%mpi_e.replace('e', '(kgen_msize-1)'), 'INT(%s)'%openmp_s.replace('e', '(kgen_osize-1)'), \
                    'INT(%s)'%openmp_e.replace('e', '(kgen_osize-1)'), 'kgen_msize', 'kgen_osize', \
                    'OMP_GET_THREAD_NUM()', 'kgen_lockpath', 'kgen_isstop']}
                part_append_gensnode(ifstop, EXEC_PART, statements.Call, attrs=attrs)

            attrs = {'expr': 'ALL(kgen_isstop)'}
            ifallstop = part_append_gensnode(ifstop, EXEC_PART, block_statements.IfThen, attrs=attrs)

            attrs = {'loopcontrol': 'kgen_count=1,3'}
            doretry = part_append_gensnode(ifallstop, EXEC_PART, block_statements.Do, attrs=attrs)

            #attrs = {'variable': 'kgen_stopunit', 'sign': '=', 'expr': 'kgen_get_newunit()'}
            #part_append_gensnode(doretry, EXEC_PART, statements.Assignment, attrs=attrs)

            attrs = {'specs': ['NEWUNIT=kgen_stopunit(OMP_GET_THREAD_NUM())', 'FILE="%s/kgen_statefile.lst"'%getinfo('kernel_path'), 'STATUS="REPLACE"', \
                'FORM="FORMATTED"', 'ACCESS="SEQUENTIAL"', 'ACTION="WRITE"', 'IOSTAT=kgen_ierr']}
            part_append_gensnode(doretry, EXEC_PART, statements.Open, attrs=attrs)

            attrs = {'expr': 'kgen_ierr .EQ. 0'}
            iflst = part_append_gensnode(doretry, EXEC_PART, block_statements.IfThen, attrs=attrs)

            attrs = {'specs': ['kgen_stopunit(OMP_GET_THREAD_NUM())']}
            part_append_gensnode(iflst, EXEC_PART, statements.Flush, attrs=attrs)

            for (mpi_s, mpi_e), (openmp_s, openmp_e), (invoke_s, invoke_e) in getinfo('invocations'):
                attrs = {'designator': 'kgen_write_list', 'items': ['kgen_stopunit(OMP_GET_THREAD_NUM())', 'INT(%s)'%mpi_s.replace('e', '(kgen_msize-1)'), \
                    'INT(%s)'%mpi_e.replace('e', '(kgen_msize-1)'), 'INT(%s)'%openmp_s.replace('e', '(kgen_osize-1)'), \
                    'INT(%s)'%openmp_e.replace('e', '(kgen_osize-1)'), 'INT(%s)'%invoke_s, 'INT(%s)'%invoke_e]}
                part_append_gensnode(iflst, EXEC_PART, statements.Call, attrs=attrs)

            attrs = {'specs': ['kgen_stopunit(OMP_GET_THREAD_NUM())']}
            part_append_gensnode(iflst, EXEC_PART, statements.Flush, attrs=attrs)

            attrs = {'specs': [ 'UNIT=kgen_stopunit(OMP_GET_THREAD_NUM())', 'STATUS="KEEP"' ]}
            part_append_gensnode(iflst, EXEC_PART, statements.Close, attrs=attrs)

            attrs = {'items': ['"Stopping application..."']}
            part_append_gensnode(iflst, EXEC_PART, statements.Write, attrs=attrs)

            attrs = {'designator': 'SLEEP', 'items': ['5']}
            part_append_gensnode(iflst, EXEC_PART, statements.Call, attrs=attrs)

            if getinfo('is_mpi_app'):
                attrs = {'designator': 'mpi_abort', 'items': [getinfo('mpi_comm'), '0', 'kgen_ierr']}
                part_append_gensnode(iflst, EXEC_PART, statements.Call, attrs=attrs)
            else:
                attrs = {'code': '0'}
                part_append_gensnode(iflst, EXEC_PART, statements.Stop, attrs=attrs)

            part_append_gensnode(iflst, EXEC_PART, statements.Else)

            attrs = {'designator': 'RANDOM_NUMBER', 'items': ['kgen_realnum']}
            part_append_gensnode(iflst, EXEC_PART, statements.Call, attrs=attrs)

            attrs = {'designator': 'SLEEP', 'items': ['INT(kgen_realnum * kgen_count * 3) + 1']}
            part_append_gensnode(iflst, EXEC_PART, statements.Call, attrs=attrs)

            #attrs = {'expr': 'OMP_IN_PARALLEL()'}
            #ifomp = namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, block_statements.IfThen, attrs=attrs)
            #part_append_comment(ifomp, EXEC_PART, 'END CRITICAL (kgen_fini)', style='openmp')
            namedpart_append_comment(node.kgen_kernel_id, BEFORE_CALLSITE, 'END CRITICAL (kgen_fini)', style='openmp')

            attrs = {'variable': 'kgen_invoke(OMP_GET_THREAD_NUM())', 'sign': '=', 'expr': 'kgen_invoke(OMP_GET_THREAD_NUM()) + 1'}
            namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, statements.Assignment, attrs=attrs)

            namedpart_append_comment(node.kgen_kernel_id, BEFORE_CALLSITE, 'END OF KGEN REGION')
        else:
            attrs = {'expr': '.NOT. kgen_ischecked(0) .AND. kgen_islast(0)'}
            ifstop = namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, block_statements.IfThen, attrs=attrs)

            attrs = {'variable': 'kgen_ischecked(0)', 'sign': '=', 'expr': '.TRUE.'}
            part_append_gensnode(ifstop, EXEC_PART, statements.Assignment, attrs=attrs)

            #attrs = {'variable': 'kgen_stopunit', 'sign': '=', 'expr': 'kgen_get_newunit()'}
            #part_append_gensnode(ifstop, EXEC_PART, statements.Assignment, attrs=attrs)

            attrs = {'specs': ['NEWUNIT=kgen_stopunit', 'FILE=TRIM(ADJUSTL(kgen_lockpath(kgen_mymid, 0)))', \
                'STATUS="NEW"', 'IOSTAT=kgen_ierr']}
            part_append_gensnode(ifstop, EXEC_PART, statements.Open, attrs=attrs)

            attrs = {'expr': 'kgen_ierr == 0'}
            ifopen = part_append_gensnode(ifstop, EXEC_PART, block_statements.IfThen, attrs=attrs)

            attrs = {'specs': [ 'UNIT=kgen_stopunit', 'STATUS="KEEP"' ]}
            part_append_gensnode(ifopen, EXEC_PART, statements.Close, attrs=attrs)
 

            for (mpi_s, mpi_e), (openmp_s, openmp_e), (invoke_s, invoke_e) in getinfo('invocations'):
                attrs = {'designator': 'kgen_check_stop', 'items': ['INT(%s)'%mpi_s.replace('e', '(kgen_msize-1)'), \
                    'INT(%s)'%mpi_e.replace('e', '(kgen_msize-1)'), 'INT(%s)'%openmp_s.replace('e', '(kgen_osize-1)'), \
                    'INT(%s)'%openmp_e.replace('e', '(kgen_osize-1)'), 'kgen_msize', 'kgen_osize', \
                    '0', 'kgen_lockpath', 'kgen_isstop']}
                part_append_gensnode(ifstop, EXEC_PART, statements.Call, attrs=attrs)

            attrs = {'expr': 'ALL(kgen_isstop)'}
            ifallstop = part_append_gensnode(ifstop, EXEC_PART, block_statements.IfThen, attrs=attrs)

            #attrs = {'variable': 'kgen_stopunit', 'sign': '=', 'expr': 'kgen_get_newunit()'}
            #part_append_gensnode(ifallstop, EXEC_PART, statements.Assignment, attrs=attrs)

            attrs = {'specs': ['NEWUNIT=kgen_stopunit', 'FILE="%s/kgen_statefile.lst"'%getinfo('kernel_path'), 'STATUS="REPLACE"', \
                'FORM="FORMATTED"', 'ACCESS="SEQUENTIAL"', 'ACTION="WRITE"', 'IOSTAT=kgen_ierr']}
            part_append_gensnode(ifallstop, EXEC_PART, statements.Open, attrs=attrs)

            attrs = {'expr': 'kgen_ierr .EQ. 0'}
            iflst = part_append_gensnode(ifallstop, EXEC_PART, block_statements.IfThen, attrs=attrs)

            attrs = {'specs': ['kgen_stopunit']}
            part_append_gensnode(iflst, EXEC_PART, statements.Flush, attrs=attrs)

            for (mpi_s, mpi_e), (openmp_s, openmp_e), (invoke_s, invoke_e) in getinfo('invocations'):
                attrs = {'designator': 'kgen_write_list', 'items': ['kgen_stopunit', 'INT(%s)'%mpi_s.replace('e', '(kgen_msize-1)'), \
                    'INT(%s)'%mpi_e.replace('e', '(kgen_msize-1)'), 'INT(%s)'%openmp_s.replace('e', '(kgen_osize-1)'), \
                    'INT(%s)'%openmp_e.replace('e', '(kgen_osize-1)'), 'INT(%s)'%invoke_s, 'INT(%s)'%invoke_e]}
                part_append_gensnode(iflst, EXEC_PART, statements.Call, attrs=attrs)

            attrs = {'specs': ['kgen_stopunit']}
            part_append_gensnode(iflst, EXEC_PART, statements.Flush, attrs=attrs)

            attrs = {'specs': [ 'UNIT=kgen_stopunit', 'STATUS="KEEP"' ]}
            part_append_gensnode(iflst, EXEC_PART, statements.Close, attrs=attrs)

            attrs = {'items': ['"Stopping application..."']}
            part_append_gensnode(iflst, EXEC_PART, statements.Write, attrs=attrs)


            if getinfo('is_mpi_app'):
                attrs = {'designator': 'SLEEP', 'items': ['5']}
                part_append_gensnode(iflst, EXEC_PART, statements.Call, attrs=attrs)

                attrs = {'designator': 'mpi_abort', 'items': [getinfo('mpi_comm'), '0', 'kgen_ierr']}
                part_append_gensnode(iflst, EXEC_PART, statements.Call, attrs=attrs)
            else:
                attrs = {'designator': 'SLEEP', 'items': ['1']}
                part_append_gensnode(iflst, EXEC_PART, statements.Call, attrs=attrs)

                attrs = {'code': '0'}
                part_append_gensnode(iflst, EXEC_PART, statements.Stop, attrs=attrs)

            attrs = {'variable': 'kgen_invoke(0)', 'sign': '=', 'expr': 'kgen_invoke(0) + 1'}
            namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, statements.Assignment, attrs=attrs)

            namedpart_append_comment(node.kgen_kernel_id, BEFORE_CALLSITE, 'END OF KGEN REGION')





"""Microapp compiler flag inspector"""

import sys
import os
import re
import io
import shutil
import time
import subprocess
import multiprocessing

from microapp import App
from fortlab.scanner.compile import kgcompiler


STR_EX = b'execve('
STR_EN = b'ENOENT'
STR_UF = b'<unfinished'
STR_RE = b'resumed>'

RE_INCLUDE = re.compile(r'(\A|^)#include\s+(\'|")(?P<name>(\w|\.)+)(\'|")\s*(!.*|)$',re.I | re.M)

class FortranCompilerOption(App):

    _name_ = "compileroption"
    _version_ = "0.1.0"

    def __init__(self, mgr):

        self.add_argument("buildcmd", metavar="build command", help="Software build command")
        self.add_argument("--cleancmd", type=str, help="Software clean command.")
        self.add_argument("--workdir", type=str, help="work directory")
        self.add_argument("--savejson", type=str, help="save data in a josn-format file")
        self.add_argument("--backupdir", type=str, help="saving source files used")
        self.add_argument("--verbose", action="store_true", help="show compilation details")
        self.add_argument("--check", action="store_true", help="check strace return code")

        self.register_forward("data", help="json object")

    def perform(self, args):

        buildcmd = args.buildcmd["_"]


        cwd = orgcwd = os.getcwd()

        if args.workdir:
            cwd = args.workdir["_"]
            os.chdir(cwd)

        if args.cleancmd:
            cleancmd_output = subprocess.check_output(args.cleancmd["_"], shell=True)

        if args.backupdir:
            backupdir = args.backupdir["_"]

        else:
            backupdir = os.path.join(os.getcwd(), "backup", "src")

        if not os.path.exists(backupdir):
            os.makedirs(backupdir)

        inq = multiprocessing.Queue()
        outq = multiprocessing.Queue()
        proc = multiprocessing.Process(target=self.get_includes, args=(backupdir, inq, outq))
        proc.start()

        # build with strace
   
        stracecmd = b'strace -f -q -s 100000 -e trace=execve -v -- /bin/sh -c "%s"'% str.encode(buildcmd)

        try:

            flags = {}

            process = subprocess.Popen(stracecmd, stdin=subprocess.PIPE, \
                        stdout=subprocess.PIPE, stderr=subprocess.PIPE, \
                        shell=True)

            while True:

                #line = process.stdout.readline()
                line = process.stderr.readline()

                if line == b'' and process.poll() is not None:
                    break

                if line:
                    pos_execve = line.find(STR_EX)
                    if pos_execve >= 0:
                        pos_enoent = line.rfind(STR_EN)
                        if pos_enoent < 0:
                            pos_last = line.rfind(STR_UF)
                            if pos_last < 0:
                                pos_last = line.rfind(b']')
                            else:
                                pos_last -= 1

                            if pos_last >= 0:
                                try:
                                    lenv = {}
                                    exec(b'exepath, cmdlist, env = %s'%line[pos_execve+len(STR_EX):(pos_last+1)], None, lenv)

                                    exepath = lenv["exepath"]
                                    cmdlist = lenv["cmdlist"]
                                    env = lenv["env"]

                                    compid = cmdlist[0].split('/')[-1]
                                    if exepath and cmdlist and compid==cmdlist[0].split('/')[-1]:
                                        compiler = kgcompiler.CompilerFactory.createCompiler(compid)
                                        if compiler:
                                            srcs, incs, macros, openmp, options = compiler.parse_option(cmdlist, self._getpwd(env))
                                            if len(srcs)>0:
                                                for src in srcs:
                                                    #flags[src] = (exepath, incs, macros, openmp, options)
                                                    flags[src] = {"compiler": exepath, "include": incs,
                                                            "macros": macros, "openmp": openmp,
                                                            "options": options, "srcbackup": []}

                                                    if backupdir:
                                                        if os.path.isfile(src):
                                                            inq.put((src, incs))

                                                        else:
                                                            print("Warning: %s is not backuped" % src)

                                                    if args.verbose:
                                                        print("Compiled: %s by %s" % (src, exepath))
                                                        print(str(options))

                                                    #if src in flags:
                                                    #    flags[src].append((exepath, incs, macros, openmp, options))
                                                    #else:
                                                    #    flags[src] = [ (exepath, incs, macros, openmp, options) ]
                                except Exception as err:
                                    raise
                                    pass

            inq.put(None)

            # get return code
            retcode = process.poll()
                                         
            if args.check and retcode != 0:
                raise Exception("strace returned non-zero value: %d" % retcode)

            backupsrcs = outq.get()

            proc.join()


        #except Exception as err:
        #    raise
        finally:
            pass

        #if args.cleancmd:
        #    cleancmd_output = subprocess.check_output(args.cleancmd["_"], shell=True)

        for fname, backups in backupsrcs.items():
            flags[fname]["srcbackup"].extend(backups)

        self.add_forward(data=flags)
#
#                    if option=='include':
#                        pathlist = Inc.get(section, option).split(':')
#                        self.config["include"]['file'][realpath]['path'].extend(pathlist)
#                    elif option in [ 'compiler', 'compiler_options' ]:
#                        self.config["include"]['file'][realpath][option] = Inc.get(section, option)
#                    else:
#                        self.config["include"]['file'][realpath]['macro'][option] = Inc.get(section, option)
#

        if args.savejson:
            jsonfile = args.savejson["_"].strip()
            dirname = os.path.dirname(jsonfile)

            if dirname and not os.path.exists(dirname):
                os.makedirs(dirname)

            #cmd = ["dict2json", "@flags", "-o", jsonfile]
            opts = ["@flags", "-o", jsonfile]
            #self.manager.run_command(cmd, forward={"flags": flags})
            ret, fwds = self.run_subapp("dict2json", opts, forward={"flags": flags})
            assert ret == 0, "dict2json returned non-zero code."

        os.chdir(orgcwd)

    def _getpwd(self, env):
        for item in env:
            if item.startswith('PWD='):
                return item[4:]
        return None

    def get_includes(self, outdir, inq, outq):

        data = {}

        srcnum = 0

        if not os.path.isdir(outdir):
            os.makedirs(outdir)

        while(True):
            if inq.empty():
                time.sleep(0.001)

            else:
                srcincs = inq.get()

                if srcincs is None:
                    break

                else:
                    self._backup(outdir, srcnum, srcincs, data)
                    srcnum += 1

        outq.put(data)

    def _backup(self, outdir, srcnum, srcincs, data):

        path, incs = srcincs

        backup = os.path.join(outdir, str(srcnum))
        shutil.copy(path, backup)
        data[path] = [backup]

        dirname = os.path.dirname(path)
        incs.insert(0, dirname)

        try:
            with io.open(path,'r', encoding="utf-8") as f:
                lines = f.read()

        except UnicodeDecodeError as err:
            with io.open(path,'r', encoding="latin-1") as f:
                lines = f.read()

        for incidx, match in enumerate(RE_INCLUDE.findall(lines)):
            incfilename = match[2].strip()

            for incdir in incs: 
                incsrc = os.path.join(incdir, incfilename)

                if os.path.isfile(incsrc):
                    incbackup = os.path.join(outdir, "%d-%d" % (srcnum, incidx))
                    shutil.copy(incsrc, incbackup)
                    data[path].append((incsrc, incbackup))
                    break



"""
Fortran type declaration statements.

-----
Permission to use, modify, and distribute this software is given under the
terms of the NumPy License. See http://scipy.org.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Author: Pearu Peterson <pearu@cens.ioc.ee>
Created: May 2006
-----
"""

__all__ = ['Integer', 'Real', 'DoublePrecision', 'Complex', 'DoubleComplex',
           #'Character', 'Logical', 'Byte', 'TypeStmt', 'Class', 'Procedure', # TEMP
           'Character', 'Logical', 'Byte', 'TypeStmt', 'Class',
           'intrinsic_type_spec', 'declaration_type_spec',
           'Implicit']

import re
import string
from fortlab.resolver.base_classes import Statement, BeginStatement, EndStatement,\
     AttributeHolder, Variable
#from utils import split_comma, AnalyzeError, name_re, is_entity_decl, is_name, CHAR_BIT, parse_array_spec
from fortlab.resolver.utils import entity_split_comma, split_comma, AnalyzeError, name_re, is_entity_decl, is_name, CHAR_BIT, parse_array_spec

from fortlab.resolver import Fortran2003 # KGEN addition
from fortlab.kgutils import traverse # KGEN additon

# Intrinsic type specification statements

class TypeDeclarationStatement(Statement):
    """
    <declaration-type-spec> [ [, <attr-spec>] :: ] <entity-decl-list>
    <declaration-type-spec> = <intrinsic-type-spec>
                              | TYPE ( <derived-type-spec> )
                              | CLASS ( <derived-type-spec> )
                              | CLASS ( * )

    <derived-type-spec> = <type-name> [ ( <type-param-spec-list> ) ]
    <type-param-spec> = [ <keyword> = ] <type-param-value>
    <type-param-value> = <scalar-int-expr> | * | :

    <intrinsic-type-spec> = INTEGER [<kind-selector>]
                            | REAL [<kind-selector>]
                            | DOUBLE PRECISION
                            | COMPLEX [<kind-selector>]
                            | CHARACTER [<char-selector>]
                            | LOGICAL [<kind-selector>]

    <kind-selector> = ( [ KIND = ] <scalar-int-initialization-expr> )
    EXTENSION:
      <kind-selector> = ( [ KIND = ] <scalar-int-initialization-expr> )
                        | * <length>

    <char-selector> = <length-selector>
                      | ( LEN = <type-param-value>, KIND = <scalar-int-initialization-expr> )
                      | ( <type-param-value>, [ KIND = ] <scalar-int-initialization-expr> )
                      | ( KIND = <scalar-int-initialization-expr> [, LEN = <type-param-value>] )
    <length-selector> = ( [ LEN = ] <type-param-value> )
                        | * <char-length> [ , ]
    <char-length> = ( <type-param-value> ) | <scalar-int-literal-constant>

    <attr-spec> = <access-spec> | ALLOCATABLE | ASYNCHRONOUS
                  | DIMENSION ( <array-spec> ) | EXTERNAL
                  | INTENT ( <intent-spec> ) | INTRINSIC
                  | <language-binding-spec> | OPTIONAL
                  | PARAMETER | POINTER | PROTECTED | SAVE
                  | TARGET | VALUE | VOLATILE
    <entity-decl> = <object-name> [ ( <array-spec> ) ] [ * <char-length> ] [ <initialization> ]
                  | <function-name> [ * <char-length> ]
    <initialization> =  = <initialization-expr>
                        | => NULL
    <access-spec> = PUBLIC | PRIVATE
    <language-binding-spec> = BIND ( C [ , NAME = <scalar-char-initialization-expr>] )
    <array-spec> =   <explicit-shape-spec-list>
                   | <assumed-shape-spec-list>
                   | <deferred-shape-spec-list>
                   | <assumed-size-spec>
    <explicit-shape-spec> = [ <lower-bound> : ] <upper-bound>
    <assumed-shape-spec> = [ <lower-bound> ] :
    <deferred-shape-spec> = :
    <assumed-size-spec> = [ <explicit-shape-spec-list> , ] [ <lower-bound> : ] *
    <bound> = <specification-expr>

    <int-literal-constant> = <digit-string> [ _ <kind-param> ]
    <digit-string> = <digit> [ <digit> ]..
    <kind-param> = <digit-string> | <scalar-int-constant-name>
    """
    f2003_class = Fortran2003.Type_Declaration_Stmt # KGEN addition

    _repr_attr_names = ['selector','attrspec','entity_decls'] + Statement._repr_attr_names


    # start of KGEN

    def tokgen(self):

        def get_edecl(node, bag, depth):
            from Fortran2003 import Entity_Decl
            if isinstance(node, Entity_Decl) and node.items[0].string==bag['name']:
                bag['edecl'] = node.tostr()
                return True

        decls = self.entity_decls
        if hasattr(self, 'new_entity_decls'):
            decls = []
            for decl in self.new_entity_decls:
                edecl = {'name': decl, 'edecl': decl}
                if hasattr(self, 'full_decl') and self.full_decl:
                    traverse(self.f2003, get_edecl, edecl)
                decls.append(edecl['edecl'])
            del new_entity_decls

        attrs = self.attrspec if hasattr(self, 'attrspec') else None
        if hasattr(self, 'new_attrspec'):
            attrs = []
            for attr in self.new_attr_specs:
                attrs.append(attr)
            del self.new_attr_specs

        if hasattr(self, 'remove_attr'):
            newattrs = []
            for attr in attrs:
                if any(attr.startswith(rattr) for rattr in self.remove_attr):
                    pass
                else:
                    newattrs.append(attr)
            attrs = newattrs
            del self.remove_attr

        s = self.tostr()
        if attrs:
            s += ', ' + ', '.join(attrs)
        if decls:
            s += ' :: ' + ', '.join(decls)
        return s

    # end of KGEN

    def process_item(self):

        item = self.item
        apply_map = item.apply_map
        clsname = self.__class__.__name__.lower()
        #line = item.get_line() # KGEN deletion
        itemline = item.get_line() # KGEN addition
        line = itemline[:] # KGEN addition
        from fortlab.resolver.block_statements import Function

        if not line.lower().startswith(clsname):
            i = 0
            j = 0
            for c in line:
                i += 1
                if c==' ': continue
                j += 1
                if j==len(clsname):
                    break
            line = line[:i].replace(' ','') + line[i:]

        assert line.lower().startswith(clsname),str(line,clsname)
        line = line[len(clsname):].lstrip()

        if line.startswith('('):
            i = line.find(')')
            selector = apply_map(line[:i+1].strip())
            line = line[i+1:].lstrip()
        elif line.startswith('*'):
            selector = '*'
            line = line[1:].lstrip()
            if line.startswith('('):
                i = line.find(')')
                selector += apply_map(line[:i+1].rstrip())
                line = line[i+1:].lstrip()
            else:
                m = re.match(r'\d+(_\w+|)|[*]',line)
                if not m:
                    self.isvalid = False
                    return
                i = m.end()
                selector += line[:i].rstrip()
                line = line[i:].lstrip()
        else:
            selector = ''

        fm = Function.match(line)
        if fm:
            l2 = line[:fm.end()]
            m2 = re.match(r'.*?\b(?P<name>\w+)\Z',l2)
            if not m2:
                self.isvalid = False
                return
            fname = m2.group('name')

            # start of KGEN deletion
            # prevending adding typedecl stmt for Function result
            # fitem = item.copy(clsname+selector+' :: '+fname,
            #                   apply_map=True)
            # self.parent.put_item(fitem)
            # end of KGEN deletion

            # start of KGEN addition
            pos = itemline.find(line)
            if pos>0:
                if not hasattr(self.parent, 'funcresult_in_stmt'):
                    self.parent.funcresult_in_stmt = {}
                self.parent.funcresult_in_stmt[fname] = self.item.apply_map(itemline[:pos].rstrip())
                self.parent.result_in_typedecl = False
            # end of KGEN addition
                
            item.clone(line)
            self.isvalid = False
            return

        if line.startswith(','):
            line = line[1:].lstrip()

        self.raw_selector = selector
        if isinstance(self, Character):
            self.selector = self._parse_char_selector(selector)
        else:
            self.selector = self._parse_kind_selector(selector)

        i = line.find('::')
        if i==-1:
            self.attrspec = []
            #self.entity_decls = split_comma(line, self.item)
            self.entity_decls = entity_split_comma(line, self.item)
        else:
            self.attrspec = split_comma(line[:i].rstrip(), self.item)
            #self.entity_decls = split_comma(line[i+2:].lstrip(), self.item)
            self.entity_decls = entity_split_comma(line[i+2:].lstrip(), self.item)
        for entity in self.entity_decls:
            if not is_entity_decl(entity):
                self.isvalid = False
                return

        if isinstance(self.parent, Function) \
               and self.parent.name in self.entity_decls:
            assert self.parent.typedecl is None,str(self.parent.typedecl)
            self.parent.typedecl = self
            self.parent.result_in_typedecl = True # KGEN addition
            #self.ignore = True # KGEN deletion
        if isinstance(self, Type):
            self.name = self.selector[1].lower()
            assert is_name(self.name),str(self.name)
        else:
            self.name = clsname
        return

    def _parse_kind_selector(self, selector):
        if not selector:
            return '',''
        length,kind = '',''
        if selector.startswith('*'):
            length = selector[1:].lstrip()
        else:
            assert selector[0]+selector[-1]=='()',str(selector)
            l = selector[1:-1].strip()
            if l.lower().startswith('kind'):
                l = l[4:].lstrip()
                if l[0]+l[-1]=='()':
                    kind = 'kind'+l
                else:
                    assert l.startswith('='),str(l)
                    kind = l[1:].lstrip()
            else:
                kind = l
        return length,kind

    def _split_char_selector(self, line):
        """line=``[key=]value`` -> key, value.
        If line does not have name part then return None, value.
        """
        for name in ['len', 'kind']:
            if line[:len(name)].lower()==name:
                value_part = line[len(name):].lstrip()
                if value_part.startswith('='):
                    return name, value_part[1:].lstrip()
        return None, line

    def _parse_char_selector(self, selector):
        if not selector:
            return '',''
        if selector.startswith('*'):
            l = selector[1:].lstrip()
            if l.startswith('('):
                if l.endswith(','): l = l[:-1].rstrip()
                assert l.endswith(')'),str(l)
                l = l[1:-1].strip()
                if l.lower().startswith('len'):
                    l = l[3:].lstrip()[1:].lstrip()
            kind=''
        else:
            assert selector[0]+selector[-1]=='()',str(selector)
            l = split_comma(selector[1:-1].strip(), self.item)
            if len(l)==1:
                l = l[0]
                key, value = self._split_char_selector(l)
                if key=='len':
                    kind, l = '', value
                elif key=='kind':
                    kind, l = value, ''
                else:
                    kind = ''
            else:
                assert len(l)==2,str(l)
                key0, value0 = self._split_char_selector(l[0])
                key1, value1 = self._split_char_selector(l[1])
                if key0=='len':
                    assert key1 in [None, 'kind'],str(key1)
                    l,kind = value0, value1
                elif key0=='kind':
                    assert key1=='len',str(key1)
                    l,kind = value1, value0
                else:
                    assert key0 is None,str(key0)
                    assert key1 in [None,'kind'],str(key1)
                    l, kind = value0, value1
        return l,kind


    # start of KGEN addition
    def _selector_str(self):
        s = ''
        length, kind = self.selector
        if isinstance(self, Character):
            if length and kind:
                s += 'LEN=%s, KIND=%s' % (length,kind)
            elif length:
                s += 'LEN=%s' % (length)
            elif kind:
                s += 'KIND=%s' % (kind)
        else:
            if isinstance(self, Type):
                s += '%s' % (kind)
            else:
                if length:
                    s += 'KIND=%s' % (length)
                if kind:
                    #if isinstance(self, (Class, Procedure)): #TEMP
                    if isinstance(self, (Class,)):
                        s += '%s' % (kind)
                    else:
                        s += 'KIND=%s' % (kind)
        return s
    # end of KGEN addition

    def tostr(self):
        # start of KGEN addition
        if hasattr(self, 'type_spec'):
            clsname = self.type_spec

            s = ''
            length, kind = self.selector if hasattr(self, 'selector') else (None, None)
            if self.type_spec.lower()=='character':
                if length and kind:
                    s += '(LEN=%s, KIND=%s)' % (length,kind)
                elif length:
                    s += '(LEN=%s)' % (length)
                elif kind:
                    s += '(KIND=%s)' % (kind)
            else:
                if self.type_spec.lower() in ['type', 'class']:
                    s += '(%s)' % (kind)
                else:
                    if length:
                        s += '*%s' % (length)
                    if kind:
                        if self.type_spec.lower()=='class':
                            s += '(%s)' % (kind)
                        else:
                            s += '(KIND=%s)' % (kind)
        else:
        # end of KGEN addition
            clsname = self.__class__.__name__.upper()

            s = ''
            length, kind = self.selector
            if isinstance(self, Character):
                if length and kind:
                    s += '(LEN=%s, KIND=%s)' % (length,kind)
                elif length:
                    s += '(LEN=%s)' % (length)
                elif kind:
                    s += '(KIND=%s)' % (kind)
            else:
                if isinstance(self, Type):
                    s += '(%s)' % (kind)
                else:
                    if length:
                        s += '*%s' % (length)
                    if kind:
                        # start of KGEN addition
                        #if isinstance(self, (Class, Procedure)):
                        if isinstance(self, (Class,)): #TEMP
                            s += '(%s)' % (kind)
                        else:
                            s += '(KIND=%s)' % (kind)
                        # end of KGEN addition
                        #s += '(KIND=%s)' % (kind) # KGEN deletion

        return clsname + s

    def tofortran(self,isfix=None):
        tab = self.get_indent_tab(isfix=isfix)
        s = self.tostr()
        if self.attrspec:
            s += ', ' + ', '.join(self.attrspec)
            #if self.entity_decls: # KGEN deletion
            #    s += ' ::' # KGEN deletion
        if self.entity_decls:
            #s += ' ' + ', '.join(self.entity_decls) # KGEN deletion
            s += ' :: ' + ', '.join(self.entity_decls) # KGEN addition
        return tab + s

    def __str__(self):
        return self.tofortran()

    def __eq__(self, other):
        if self.__class__ is not other.__class__:
            return False
        return self.selector==other.selector

    def astypedecl(self):
        if self.entity_decls or self.attrspec:
            return self.__class__(self.parent, self.item.copy(self.tostr()))
        return self

    def analyze(self):
        if not self.entity_decls:
            return
        variables = self.parent.a.variables
        typedecl = self.astypedecl()
        attrspec = self.attrspec[:]
        access_spec_lst = [a for a in attrspec if a.lower() in ['private','public']]
        if access_spec_lst:
            access_spec = access_spec_lst[0]
            attrspec.remove(access_spec)
        else:
            access_spec = None
        for item in self.entity_decls:
            name, array_spec, char_length, value = self._parse_entity(item)
            var = self.parent.get_variable(name)
            var.add_parent(self)
            if char_length:
                var.set_length(char_length)
            var.set_type(typedecl)
            var.update(self.attrspec)
            if array_spec:
                var.set_bounds(array_spec)
            if value:
                var.set_init(value)
            if access_spec is not None:
                l = getattr(self.parent.a,access_spec.lower() + '_id_list')
                l.append(name)
            var.analyze()
        return

    def _parse_entity(self, line):
        m = name_re(line)
        assert m,str(line,self.item,self.__class__.__name__)
        name = line[:m.end()]
        line = line[m.end():].lstrip()
        array_spec = None
        char_length = None
        value = None
        if line:
            item = self.item.copy(line)
            line = item.get_line()
            if line.startswith('('):
                i = line.find(')')
                assert i!=-1,str(line)
                array_spec = parse_array_spec(line[1:i].strip(), item)
                line = line[i+1:].lstrip()

            if line.startswith('*'):
                i = line.find('=')
                if i==-1:
                    char_length = item.apply_map(line[1:].lstrip())
                    line = ''
                else:
                    char_length = item.apply_map(line[1:i].strip())
                    line = line[i:]
            if line.startswith('='):
                value = item.apply_map(line[1:].lstrip())
        return name, array_spec, char_length, value

    def get_zero_value(self):
        raise NotImplementedError(str(self.__class__.__name__))

    def assign_expression(self, name, value):
        return '%s = %s' % (name, value)

    def get_kind(self):
        return self.selector[1] or self.default_kind

    def get_length(self):
        return self.selector[0] or 1

    def get_byte_size(self):
        length, kind = self.selector
        if length: return int(length)
        if kind: return int(kind)
        return self.default_kind

    def get_bit_size(self):
        return CHAR_BIT * int(self.get_byte_size())

    #def is_intrinsic(self): return not isinstance(self,(Type,Class,Procedure)) # KGEN deletion
    #def is_intrinsic(self): return not isinstance(self,(Type,Class,Procedure)) # KGEN addition # TEMP
    def is_intrinsic(self): return not isinstance(self,(Type,Class,)) # KGEN addition
    def is_derived(self): return isinstance(self,Type)
    def is_class(self): return isinstance(self,Class) # KGEN addition
    #def is_procedure(self): return isinstance(self,Procedure) # KGEN addition # TEMP
    def is_procedure(self): return isinstance(self,) # KGEN addition

    def is_numeric(self): return isinstance(self,(Integer,Real, DoublePrecision,Complex,DoubleComplex,Byte))
    def is_nonnumeric(self): return isinstance(self,(Character,Logical))

class Integer(TypeDeclarationStatement):
    match = re.compile(r'integer\b',re.I).match
    default_kind = 4

    def get_zero_value(self):
        kind = self.get_kind()
        if kind==self.default_kind: return '0'
        return '0_%s' % (kind)

class Real(TypeDeclarationStatement):
    match = re.compile(r'real\b',re.I).match
    default_kind = 4

    def get_zero_value(self):
        kind = self.get_kind()
        if kind==self.default_kind: return '0.0'
        return '0_%s' % (kind)

class DoublePrecision(TypeDeclarationStatement):
    match = re.compile(r'double\s*precision\b',re.I).match
    default_kind = 8

    def get_byte_size(self):
        return self.default_kind

    def get_zero_value(self):
        return '0.0D0'

class Complex(TypeDeclarationStatement):
    match = re.compile(r'complex\b',re.I).match
    default_kind = 4

    def get_byte_size(self):
        length, kind = self.selector
        if length: return int(length)
        if kind: return 2*int(kind)
        return 2*self.default_kind

    def get_zero_value(self):
        kind = self.get_kind()
        if kind==self.default_kind: return '(0.0, 0.0)'
        return '(0.0_%s, 0.0_%s)' % (kind, kind)

    def get_part_typedecl(self):
        bz = self.get_byte_size()/2
        return Real(self.parent, self.item.copy('REAL*%s' % (bz)))

class DoubleComplex(TypeDeclarationStatement):
    # not in standard
    match = re.compile(r'double\s*complex\b',re.I).match
    default_kind = 8

    def get_byte_size(self):
        return 2*self.default_kind

    def get_zero_value(self):
        return '(0.0D0,0.0D0)'

class Logical(TypeDeclarationStatement):
    match = re.compile(r'logical\b',re.I).match
    default_kind = 4

    def get_zero_value(self):
        return ".FALSE."

class Character(TypeDeclarationStatement):
    match = re.compile(r'character\b',re.I).match
    default_kind = 1

    def get_bit_size(self):
        length = self.get_length()
        if length=='*':
            return 0  # model for character*(*)
        return CHAR_BIT * int(length) * int(self.get_kind())

    def get_zero_value(self):
        return "''"

class Byte(TypeDeclarationStatement):
    # not in standard
    match = re.compile(r'byte\b',re.I).match
    default_kind = 1

    def get_zero_value(self):
        return '0'

class Type(TypeDeclarationStatement):
    match = re.compile(r'type\s*\(', re.I).match

    def get_zero_value(self):
        type_decl = self.get_type_decl(self.name)
        component_names = type_decl.a.component_names
        components = type_decl.a.components
        l = []
        for name in component_names:
            var = components[name]
            l.append(var.typedecl.get_zero_value())
        return '%s(%s)' % (type_decl.name, ', '.join(l))

    def get_kind(self):
        # See 4.5.2, page 48
        raise NotImplementedError(str(self.__class__.__name__))

    def get_bit_size(self):
        return self.get_type_decl(self.name).get_bit_size()

TypeStmt = Type

class Class(TypeDeclarationStatement):
    match = re.compile(r'class\s*\(', re.I).match

# start of KGEN addition
#class Procedure(TypeDeclarationStatement):
#    f2003_class = Fortran2003.Procedure_Declaration_Stmt # KGEN addition
#    match = re.compile(r'procedure\s*\(', re.I).match
# end of KGEN addition

class Implicit(Statement):
    """
    IMPLICIT <implicit-spec-list>
    IMPLICIT NONE
    <implicit-spec> = <declaration-type-spec> ( <letter-spec-list> )
    <letter-spec> = <letter> [ - <letter> ]
    """
    f2003_class = Fortran2003.Implicit_Stmt # KGEN addition

    match = re.compile(r'implicit\b',re.I).match

    #letters = string.lowercase
    letters = string.ascii_lowercase

    def process_item(self):
        line = self.item.get_line()[8:].lstrip()
        if line.lower()=='none':
            self.items = []
            return
        items = []
        for item in split_comma(line, self.item):
            i = item.find('(')
            assert i!=-1 and item.endswith(')'),str(item)
            specs = []
            for spec in split_comma(item[i+1:-1].strip(), self.item):
                if '-' in spec:
                    s,e = spec.lower().split('-')
                    s = s.strip()
                    e = e.strip()
                    assert s in self.letters and e in self.letters,str(s)+str(e)
                else:
                    e = s = spec.lower().strip()
                    assert s in self.letters,str(s)+str(e)
                specs.append((s,e))
            tspec = item[:i].rstrip()
            stmt = None
            for cls in declaration_type_spec:
                if cls.match(tspec):
                    stmt = cls(self, self.item.copy(tspec))
                    if stmt.isvalid:
                        break
            assert stmt is not None,str(item)+str(line)
            items.append((stmt,specs))
        self.items = items
        return

    def tofortran(self, isfix=None):
        tab = self.get_indent_tab(isfix=isfix)
        if not self.items:
            return tab + 'IMPLICIT NONE'
        l = []
        for stmt,specs in self.items:
            l1 = []
            for s,e in specs:
                if s==e:
                    l1.append(s)
                else:
                    l1.append(s + '-' + e)
            l.append('%s ( %s )' % (stmt.tostr(), ', '.join(l1)))
        return tab + 'IMPLICIT ' + ', '.join(l)

    def analyze(self):
        # start of KGEN addition
        if not hasattr(self.parent, 'spec_stmts'):
            self.parent.spec_stmts = []
        self.parent.spec_stmts.append(self)
        # end of KGEN addition

        implicit_rules = self.parent.a.implicit_rules
        if not self.items:
            if implicit_rules:
                self.warning('overriding previously set implicit rule mapping'\
                      ' %r.' % (implicit_rules))
            self.parent.a.implicit_rules = None
            return
        if implicit_rules is None:
            self.warning('overriding previously set IMPLICIT NONE')
            self.parent.a.implicit_rules = implicit_rules = {}
        for stmt,specs in self.items:
            for s,e in specs:
                for l in string.lowercase[string.lowercase.index(s.lower()):\
                                          string.lowercase.index(e.lower())+1]:
                    implicit_rules[l] = stmt
        return

    # start of KGEN addition
    def tokgen(self):
        if not hasattr(self, 'items') or not self.items:
            return 'IMPLICIT NONE'
        l = []
        for stmt,specs in self.items:
            l1 = []
            for s,e in specs:
                if s==e:
                    l1.append(s)
                else:
                    l1.append(s + '-' + e)
            l.append('%s ( %s )' % (stmt.tostr(), ', '.join(l1)))
        return 'IMPLICIT ' + ', '.join(l)

    def resolve_uname(self, uname, request, config):
        self.add_geninfo(uname, request)
        if self.items:
            raise ProgramException('Custom Implicit Rule is not supported yet.')
    # end of KGEN addition

intrinsic_type_spec = [ \
    Integer , Real,
    DoublePrecision, Complex, DoubleComplex, Character, Logical, Byte
    ]
#declaration_type_spec = intrinsic_type_spec + [ TypeStmt, Class, Procedure ] # TEMP
declaration_type_spec = intrinsic_type_spec + [ TypeStmt, Class ]



# gen_write_typedecl_in_type.py
from __future__ import absolute_import
 
from fortlab.resolver import statements, block_statements, typedecl_statements
from fortlab.kgplugin import Kgen_Plugin

from .gencore_utils import get_dtype_writename, get_typedecl_writename, state_gencore_contains, \
    get_dtype_readname, get_typedecl_readname, kernel_gencore_contains, gen_read_istrue, gen_write_istrue, \
    is_zero_array, is_excluded, is_remove_state

class Gen_Typedecl_In_Type(Kgen_Plugin):
    def __init__(self):
        self.frame_msg = None
        self.state_created_subrs = []
        self.kernel_created_subrs = []

    # registration
    def register(self, msg):
        self.frame_msg = msg

        # register initial events
        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.STATE, GENERATION_STAGE.BEGIN_PROCESS, \
            typedecl_statements.TypeDeclarationStatement, self.is_typedecl_in_type, self.create_subr_write_typedecl_in_type) 

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.BEGIN_PROCESS, \
            typedecl_statements.TypeDeclarationStatement, self.is_typedecl_in_type, self.create_subr_read_typedecl_in_type) 

    def is_typedecl_in_type(self, node):
        parent = node.kgen_parent
        if parent.kgen_match_class is block_statements.Type and parent.kgen_stmt and node.kgen_stmt and \
            hasattr(parent.kgen_stmt, 'geninfo') and len(parent.kgen_stmt.geninfo)>0: 
            return True
        else: return False

    def create_read_subr(self, subrname, entity_name, parent, var, stmt, allocate=False):
        checks = lambda n: isinstance(n.kgen_stmt, block_statements.Subroutine) and n.name==subrname
        if subrname not in self.kernel_created_subrs and not part_has_node(parent, SUBP_PART, checks):

            if is_remove_state(entity_name, stmt): return

            self.kernel_created_subrs.append(subrname)

            checks = lambda n: n.kgen_isvalid and n.kgen_match_class==statements.Contains
            if not parent in kernel_gencore_contains and not part_has_node(parent, CONTAINS_PART, checks):
                part_append_comment(parent, CONTAINS_PART, '')
                part_append_genknode(parent, CONTAINS_PART, statements.Contains)
                part_append_comment(parent, CONTAINS_PART, '')
                kernel_gencore_contains.append(parent)

            part_append_comment(parent, SUBP_PART, 'write state subroutine for %s'%subrname)
            attrs = {'name': subrname, 'args': ['var', 'kgen_unit', 'printname', 'printvar']}
            subrobj = part_append_genknode(parent, SUBP_PART, block_statements.Subroutine, attrs=attrs)
            part_append_comment(parent, SUBP_PART, '')

            # variable A
            #import pdb; pdb.set_trace()
            attrspec = ['INTENT(INOUT)']
            if var.is_allocatable() or allocate: attrspec.append('ALLOCATABLE')
            if var.is_pointer(): attrspec.append('POINTER')
            if var.is_array(): attrspec.append('DIMENSION(%s)'% ','.join(':'*var.rank))
            attrs = {'type_spec': stmt.__class__.__name__.upper(), 'attrspec': attrspec, 'selector':stmt.selector, 'entity_decls': ['var']}
            part_append_genknode(subrobj, DECL_PART, stmt.__class__, attrs=attrs)

            # kgen_unit
            attrs = {'type_spec': 'INTEGER', 'attrspec': ['INTENT(IN)'], 'entity_decls': ['kgen_unit']}
            part_append_genknode(subrobj, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            # printname
            attrs = {'type_spec': 'CHARACTER', 'attrspec': ['INTENT(IN)'], 'selector':('*', None), 'entity_decls': ['printname']}
            part_append_genknode(subrobj, DECL_PART, typedecl_statements.Character, attrs=attrs)

            # printvar
            attrs = {'type_spec': 'LOGICAL', 'attrspec': ['INTENT(IN)', 'OPTIONAL'], 'selector':(None, None), 'entity_decls': ['printvar']}
            part_append_genknode(subrobj, DECL_PART, typedecl_statements.Character, attrs=attrs)

            # kgen_istrue
            attrs = {'type_spec': 'LOGICAL', 'entity_decls': ['kgen_istrue']}
            part_append_genknode(subrobj, DECL_PART, typedecl_statements.Logical, attrs=attrs)

            attrs = {'type_spec': 'REAL', 'entity_decls': ['kgen_array_sum'], 'selector': (None, '8')}
            part_append_genknode(subrobj, DECL_PART, typedecl_statements.Real, attrs=attrs)

            # array index A
            if var.is_array():
                attrs = {'type_spec': 'INTEGER', 'entity_decls': [ 'idx%d'%(d+1) for d in range(var.rank) ]}
                part_append_genknode(subrobj, DECL_PART, typedecl_statements.Integer, attrs=attrs)

                attrs = {'type_spec': 'INTEGER', 'attrspec': ['DIMENSION(2,%d)'%var.rank], 'entity_decls': [ 'kgen_bound' ]}
                part_append_genknode(subrobj, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            part_append_comment(subrobj, DECL_PART, '')

            pobj = gen_read_istrue(subrobj, var, 'var')
            part_append_comment(subrobj, EXEC_PART, '')

            if var.is_array():

                bound_args = []
                for dim in range(var.rank):
                    attrs = {'items': ['kgen_bound(1, %d)'%(dim+1)], 'specs': ['UNIT = kgen_unit']}
                    part_append_genknode(pobj, EXEC_PART, statements.Read, attrs=attrs)

                    attrs = {'items': ['kgen_bound(2, %d)'%(dim+1)], 'specs': ['UNIT = kgen_unit']}
                    part_append_genknode(pobj, EXEC_PART, statements.Read, attrs=attrs)

                    #bound_args.append('kgen_bound(2,%d)-kgen_bound(1,%d)+1'%(dim+1, dim+1))
                    bound_args.append('kgen_bound(1,%d):kgen_bound(2,%d)'%(dim+1, dim+1))

                if var.is_allocatable() or var.is_pointer() or allocate:
                    attrs = {'items': ['var(%s)'%', '.join(bound_args)]}
                    part_append_genknode(pobj, EXEC_PART, statements.Allocate, attrs=attrs)

                if stmt.is_derived():
                    indexes = [ 'idx%d'%(d+1) for d in range(var.rank) ]
                    str_indexes = ','.join(indexes)
                    #tostr_indexes = ','.join([ 'kgen_tostr(idx)' for idx in indexes])

                    prevobj = pobj
                    doobjs = []
                    for d in range(var.rank):
                        attrs = {'loopcontrol': 'idx%(d)d=kgen_bound(1,%(d)d), kgen_bound(2,%(d)d)'%{'d':d+1}}
                        doobj = part_append_genknode(prevobj, EXEC_PART, block_statements.Do, attrs=attrs)

                        doobjs.append(doobj)
                        prevobj = doobj

                    attrs = {'expr': 'PRESENT( printvar ) .AND. printvar'}
                    ifpvarobj = part_append_genknode(doobjs[-1], EXEC_PART, block_statements.IfThen, attrs=attrs)

                    callname = None
                    for uname, req in stmt.unknowns.items():
                        if uname.firstpartname()==stmt.name and len(req.res_stmts)>0:
                            res = req.res_stmts[0]
                            callname = get_dtype_readname(res)
                            break
                    if callname is None:
                        print('WARNING: Can not find Type resolver for %s'%stmt.name)
                        part_append_comment(ifpvarobj, EXEC_PART, \
                            'ERROR: "%s" is not resolved. Call statements to read "%s" is not created here.'%\
                            (stmt.name, stmt.name))
                    else:
                        attrs = {'designator': callname, 'items': ['var(%s)'%str_indexes, 'kgen_unit', 'printname // "(%s)"'%str_indexes, '.TRUE.']}
                        part_append_genknode(ifpvarobj, EXEC_PART, statements.Call, attrs=attrs)

                        part_append_genknode(ifpvarobj, EXEC_PART, statements.Else)

                        pstr = '.TRUE.' if any(match_namepath(pattern, pack_exnamepath(stmt, entity_name), internal=False) for pattern in getinfo('print_var_names')) else '.FALSE.'
                        attrs = {'designator': callname, 'items': ['var(%s)'%str_indexes, 'kgen_unit', 'printname // "(%s)"'%str_indexes, pstr]}
                        part_append_genknode(ifpvarobj, EXEC_PART, statements.Call, attrs=attrs)

                else: # intrinsic type
                    attrs = {'items': ['var'], 'specs': ['UNIT = kgen_unit']}
                    part_append_genknode(pobj, EXEC_PART, statements.Read, attrs=attrs)

                    if stmt.is_numeric():
                        attrs = {'designator': 'kgen_array_sumcheck', 'items': ['printname', 'kgen_array_sum', 'DBLE(SUM(var, mask=(var .eq. var)))', '.TRUE.']}
                        part_append_genknode(pobj, EXEC_PART, statements.Call, attrs=attrs)

                    attrs = {'expr': 'PRESENT( printvar ) .AND. printvar'}
                    ifpvarobj = part_append_genknode(pobj, EXEC_PART, block_statements.IfThen, attrs=attrs)
                    if stmt.is_numeric():
                        attrs = {'items': ['"KGEN DEBUG: DBLE(SUM(" // printname // ")) = "', 'DBLE(SUM(var, mask=(var .eq. var)))']}
                    else:
                        attrs = {'items': ['"KGEN DEBUG: " // printname // " = "', 'var']}
                    part_append_genknode(ifpvarobj, EXEC_PART, statements.Write, attrs=attrs)

                    if any(match_namepath(pattern, pack_exnamepath(stmt, entity_name), internal=False) for pattern in getinfo('print_var_names')):

                        part_append_genknode(ifpvarobj, EXEC_PART, statements.Else)

                        if stmt.is_numeric():
                            attrs = {'items': ['"KGEN DEBUG: DBLE(SUM(" // printname // ")) = "', 'DBLE(SUM(var, mask=(var .eq. var)))']}
                        else:
                            attrs = {'items': ['"KGEN DEBUG: " // printname // " = "', 'var']}
                        part_append_genknode(ifpvarobj, EXEC_PART, statements.Write, attrs=attrs)

            else: # scalar
                if var.is_allocatable() or var.is_pointer() or allocate:
                    attrs = {'items': ['var']}
                    part_append_genknode(pobj, EXEC_PART, statements.Allocate, attrs=attrs)

                if stmt.is_derived():
                    attrs = {'expr': 'PRESENT( printvar ) .AND. printvar'}
                    ifpvarobj = part_append_genknode(pobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    callname = None
                    for uname, req in stmt.unknowns.items():
                        if uname.firstpartname()==stmt.name and len(req.res_stmts)>0:
                            res = req.res_stmts[0]
                            callname = get_dtype_readname(res)
                            break
                    if callname is None:
                        print('WARNING: Can not find Type resolver for %s'%stmt.name)
                        part_append_comment(ifpvarobj, EXEC_PART, \
                            'ERROR: "%s" is not resolved. Call statements to read "%s" is not created here.'%\
                            (stmt.name, stmt.name))
                    else:
                        attrs = {'designator': callname, 'items': ['var', 'kgen_unit', 'printname', '.TRUE.']}
                        part_append_genknode(ifpvarobj, EXEC_PART, statements.Call, attrs=attrs)

                        part_append_genknode(ifpvarobj, EXEC_PART, statements.Else)

                        pstr = '.TRUE.' if any(match_namepath(pattern, pack_exnamepath(stmt, entity_name), internal=False) for pattern in getinfo('print_var_names')) else '.FALSE.'
                        attrs = {'designator': callname, 'items': ['var', 'kgen_unit', 'printname', pstr]}
                        part_append_genknode(ifpvarobj, EXEC_PART, statements.Call, attrs=attrs)

                else: # intrinsic type
                    attrs = {'items': ['var'], 'specs': ['UNIT = kgen_unit']}
                    part_append_genknode(pobj, EXEC_PART, statements.Read, attrs=attrs)

                    attrs = {'expr': 'PRESENT( printvar )  .AND. printvar'}
                    ifpvarobj = part_append_genknode(pobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    attrs = {'items': ['"KGEN DEBUG: " // printname // " = "', 'var']}
                    part_append_genknode(ifpvarobj, EXEC_PART, statements.Write, attrs=attrs)

                    if any(match_namepath(pattern, pack_exnamepath(stmt, entity_name), internal=False) for pattern in getinfo('print_var_names')):

                        part_append_genknode(ifpvarobj, EXEC_PART, statements.Else)

                        attrs = {'items': ['"KGEN DEBUG: " // printname // " = "', 'var']}
                        part_append_genknode(ifpvarobj, EXEC_PART, statements.Write, attrs=attrs)

    def create_write_subr(self, subrname, entity_name, parent, var, stmt):

        checks = lambda n: isinstance(n.kgen_stmt, block_statements.Subroutine) and n.name==subrname
        if subrname not in self.state_created_subrs and not part_has_node(parent, SUBP_PART, checks):

            if is_remove_state(entity_name, stmt): return

            self.state_created_subrs.append(subrname)

            checks = lambda n: n.kgen_match_class==statements.Contains
            if not parent in state_gencore_contains and not part_has_node(parent, CONTAINS_PART, checks):
                part_append_comment(parent, CONTAINS_PART, '')
                part_append_gensnode(parent, CONTAINS_PART, statements.Contains)
                part_append_comment(parent, CONTAINS_PART, '')
                state_gencore_contains.append(parent)

            part_append_comment(parent, SUBP_PART, 'write state subroutine for %s'%subrname)
            attrs = {'name': subrname, 'args': ['var', 'kgen_unit', 'printname', 'printvar']}
            subrobj = part_append_gensnode(parent, SUBP_PART, block_statements.Subroutine, attrs=attrs)
            part_append_comment(parent, SUBP_PART, '')

            parent.kgen_stmt.top.used4genstate = True

            # variable A
            #import pdb; pdb.set_trace()
            attrspec = ['INTENT(IN)']
            if var.is_allocatable(): attrspec.append('ALLOCATABLE')
            if var.is_pointer(): attrspec.append('POINTER')
            if var.is_array(): attrspec.append('DIMENSION(%s)'% ','.join(':'*var.rank))
            attrs = {'type_spec': stmt.__class__.__name__.upper(), 'attrspec': attrspec, 'selector':stmt.selector, 'entity_decls': ['var']}
            part_append_gensnode(subrobj, DECL_PART, stmt.__class__, attrs=attrs)

            # kgen_unit
            attrs = {'type_spec': 'INTEGER', 'attrspec': ['INTENT(IN)'], 'entity_decls': ['kgen_unit']}
            part_append_gensnode(subrobj, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            # printname
            attrs = {'type_spec': 'CHARACTER', 'attrspec': ['INTENT(IN)'], 'selector':('*', None), 'entity_decls': ['printname']}
            part_append_gensnode(subrobj, DECL_PART, typedecl_statements.Character, attrs=attrs)

            # printvar
            attrs = {'type_spec': 'LOGICAL', 'attrspec': ['INTENT(IN)', 'OPTIONAL'], 'selector':(None, None), 'entity_decls': ['printvar']}
            part_append_gensnode(subrobj, DECL_PART, typedecl_statements.Character, attrs=attrs)

            # kgen_istrue
            attrs = {'type_spec': 'LOGICAL', 'entity_decls': ['kgen_istrue']}
            part_append_gensnode(subrobj, DECL_PART, typedecl_statements.Logical, attrs=attrs)

            # array index A
            if var.is_array():
                attrs = {'type_spec': 'INTEGER', 'entity_decls': [ 'idx%d'%(d+1) for d in range(var.rank) ]}
                part_append_gensnode(subrobj, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            attrs = {'type_spec': 'REAL', 'entity_decls': ['kgen_array_sum'], 'selector': (None, '8')}
            part_append_genknode(subrobj, DECL_PART, typedecl_statements.Real, attrs=attrs)

            part_append_comment(subrobj, DECL_PART, '')

            pobj = gen_write_istrue(subrobj, var, 'var')

            if var.is_array():

                for dim in range(var.rank):
                    attrs = {'items': ['LBOUND(var, %d)'%(dim+1)], 'specs': ['UNIT = kgen_unit']}
                    part_append_gensnode(pobj, EXEC_PART, statements.Write, attrs=attrs)

                    attrs = {'items': ['UBOUND(var, %d)'%(dim+1)], 'specs': ['UNIT = kgen_unit']}
                    part_append_gensnode(pobj, EXEC_PART, statements.Write, attrs=attrs)

                if stmt.is_derived():
                    indexes = [ 'idx%d'%(d+1) for d in range(var.rank) ]
                    str_indexes = ','.join(indexes)
                    #tostr_indexes = ','.join([ 'kgen_tostr(idx)' for idx in indexes])

                    prevobj = pobj
                    doobjs = []
                    for d in range(var.rank):
                        attrs = {'loopcontrol': 'idx%(d)d=LBOUND(var,%(d)d), UBOUND(var,%(d)d)'%{'d':d+1}}
                        doobj = part_append_gensnode(prevobj, EXEC_PART, block_statements.Do, attrs=attrs)

                        doobjs.append(doobj)
                        prevobj = doobj

                    attrs = {'expr': 'PRESENT( printvar ) .AND. printvar'}
                    ifpvarobj = part_append_gensnode(doobjs[-1], EXEC_PART, block_statements.IfThen, attrs=attrs)

                    callname = None
                    for uname, req in stmt.unknowns.items():
                        if uname.firstpartname()==stmt.name and len(req.res_stmts)>0:
                            res = req.res_stmts[0]
                            callname = get_dtype_writename(res)
                            break
                    if callname is None:
                        print('WARNING: Can not find Type resolver for %s'%stmt.name)
                        part_append_comment(ifpvarobj, EXEC_PART, \
                            'ERROR: "%s" is not resolved. Call statements to write "%s" is not created here.'%\
                            (stmt.name, stmt.name))
                    else:
                        attrs = {'designator': callname, 'items': ['var(%s)'%str_indexes, 'kgen_unit', 'printname // "(%s)"'%str_indexes, '.TRUE.']}
                        part_append_gensnode(ifpvarobj, EXEC_PART, statements.Call, attrs=attrs)

                        part_append_gensnode(ifpvarobj, EXEC_PART, statements.Else)

                        pstr = '.TRUE.' if any(match_namepath(pattern, pack_exnamepath(stmt, entity_name), internal=False) for pattern in getinfo('print_var_names')) else '.FALSE.'
                        attrs = {'designator': callname, 'items': ['var(%s)'%str_indexes, 'kgen_unit', 'printname // "(%s)"'%str_indexes, pstr]}
                        part_append_gensnode(ifpvarobj, EXEC_PART, statements.Call, attrs=attrs)

                else: # intrinsic type
                    attrs = {'items': ['var'], 'specs': ['UNIT = kgen_unit']}
                    part_append_gensnode(pobj, EXEC_PART, statements.Write, attrs=attrs)

                    attrs = {'expr': 'PRESENT( printvar ) .AND. printvar'}
                    ifpvarobj = part_append_gensnode(pobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    if stmt.is_numeric():
                        attrs = {'items': ['"KGEN DEBUG: DBLE(SUM(" // printname // ") = "', 'DBLE(SUM(var, mask=(var .eq. var)))']}
                    else:
                        attrs = {'items': ['"KGEN DEBUG: " // printname // " = "', 'var']}
                    part_append_gensnode(ifpvarobj, EXEC_PART, statements.Write, attrs=attrs)

                    if any(match_namepath(pattern, pack_exnamepath(stmt, entity_name), internal=False) for pattern in getinfo('print_var_names')):

                        part_append_gensnode(ifpvarobj, EXEC_PART, statements.Else)

                        if stmt.is_numeric():
                            attrs = {'items': ['"KGEN DEBUG: " // "DBLE(SUM(" // printname // ")) = "', 'DBLE(SUM(var, mask=(var .eq. var)))']}
                        else:
                            attrs = {'items': ['"KGEN DEBUG: " // printname // " = "', 'var']}
                        part_append_gensnode(ifpvarobj, EXEC_PART, statements.Write, attrs=attrs)

            else: # scalar
                if stmt.is_derived():
                    attrs = {'expr': 'PRESENT( printvar ) .AND. printvar'}
                    ifpvarobj = part_append_gensnode(pobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    callname = None
                    for uname, req in stmt.unknowns.items():
                        if uname.firstpartname()==stmt.name and len(req.res_stmts)>0:
                            res = req.res_stmts[0]
                            callname = get_dtype_writename(res)
                            break
                    if callname is None:
                        print('WARNING: Can not find Type resolver for %s'%stmt.name)
                        part_append_comment(ifpvarobj, EXEC_PART, \
                            'ERROR: "%s" is not resolved. Call statements to write "%s" is not created here.'%\
                            (stmt.name, stmt.name))
                    else:
                        attrs = {'designator': callname, 'items': ['var', 'kgen_unit', 'printname', '.TRUE.']}
                        part_append_gensnode(ifpvarobj, EXEC_PART, statements.Call, attrs=attrs)

                        part_append_gensnode(ifpvarobj, EXEC_PART, statements.Else)

                        pstr = '.TRUE.' if any(match_namepath(pattern, pack_exnamepath(stmt, entity_name), internal=False) for pattern in getinfo('print_var_names')) else '.FALSE.'
                        attrs = {'designator': callname, 'items': ['var', 'kgen_unit', 'printname', pstr]}
                        part_append_gensnode(ifpvarobj, EXEC_PART, statements.Call, attrs=attrs)

                else: # intrinsic type
                    attrs = {'items': ['var'], 'specs': ['UNIT = kgen_unit']}
                    part_append_gensnode(pobj, EXEC_PART, statements.Write, attrs=attrs)

                    attrs = {'expr': 'PRESENT( printvar ) .AND. printvar'}
                    ifpvarobj = part_append_gensnode(pobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    attrs = {'items': ['"KGEN DEBUG: " // printname // " = "', 'var']}
                    part_append_gensnode(ifpvarobj, EXEC_PART, statements.Write, attrs=attrs)

                    if any(match_namepath(pattern, pack_exnamepath(stmt, entity_name), internal=False) for pattern in getinfo('print_var_names')):

                        part_append_gensnode(ifpvarobj, EXEC_PART, statements.Else)

                        attrs = {'items': ['"KGEN DEBUG: " // printname // " = "', 'var']}
                        part_append_gensnode(ifpvarobj, EXEC_PART, statements.Write, attrs=attrs)

    def create_subr_read_typedecl_in_type(self, node):

        stmt = node.kgen_stmt
        parent = node.kgen_parent.kgen_parent
        entity_names = [ get_entity_name(decl) for decl in stmt.entity_decls ]

        for entity_name, entity_decl in zip(entity_names, stmt.entity_decls):
            var = stmt.get_variable(entity_name)
            subrname = get_typedecl_readname(stmt, entity_name)
            if subrname is None: raise Exception('Can not get subroutinename')

            if var.is_array():
                if is_zero_array(var, stmt): continue

                if stmt.is_derived():
                    self.create_read_subr(subrname, entity_name, parent, var, stmt)
                else: # intrinsic type
                    if var.is_explicit_shape_array():
                        pass
                    else: # implicit array
                        self.create_read_subr(subrname, entity_name, parent, var, stmt)
            else:
                if stmt.is_derived():
                    if var.is_allocatable() or var.is_pointer():
                        self.create_read_subr(subrname, entity_name, parent, var, stmt)
     
    def create_subr_write_typedecl_in_type(self, node):

        stmt = node.kgen_stmt
        parent = node.kgen_parent.kgen_parent
        entity_names = [ get_entity_name(decl) for decl in stmt.entity_decls ]

        for entity_name, entity_decl in zip(entity_names, stmt.entity_decls):
            var = stmt.get_variable(entity_name)
            subrname = get_typedecl_writename(stmt, entity_name)
            if subrname is None: raise Exception('Can not get subroutinename')

            if var.is_array():
                if is_zero_array(var, stmt): continue

                if stmt.is_derived():
                    self.create_write_subr(subrname, entity_name, parent, var, stmt)
                else: # intrinsic type
                    if var.is_explicit_shape_array():
                        pass
                    else: # implicit array
                        self.create_write_subr(subrname, entity_name, parent, var, stmt)
            else:
                if stmt.is_derived():
                    if var.is_allocatable() or var.is_pointer():
                        self.create_write_subr(subrname, entity_name, parent, var, stmt)



# simple_timing.py

#import os 
from fortlab.resolver import statements, block_statements, typedecl_statements
from fortlab.kgplugin import Kgen_Plugin

KERNEL_PBLOCK_TIMING = 'KPBTIM'

class Simple_Timing(Kgen_Plugin):
    def __init__(self):
        self.frame_msg = None

    # registration
    def register(self, msg):
        self.frame_msg = msg

        # register event per function 
        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.NODE_CREATED, \
            getinfo('parentblock_stmt'), None, self.register_event) 

    def ispstmt(self, stmt, cmpstmt, limitstmt):
        if stmt == cmpstmt:
            return True
        elif stmt == limitstmt:
            return False
        elif hasattr(stmt, 'parent'):
            return self.ispstmt(stmt.parent, cmpstmt, limitstmt)
        else:
            return False

    def register_event(self, node):

#        if getinfo('is_papi_enabled'):
#            part_append_comment(node, DECL_PART, '#ifdef KGEN_PAPI', style='rawtext')
#
#            attrs = {'type_spec': 'INTEGER', 'attrspec': ['DIMENSION(1)'], \
#                'entity_decls': ['kgen_papi_events = (/ KGENPAPIEVENT /)']}
#            part_append_genknode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)
#
#            attrs = {'type_spec': 'INTEGER', 'attrspec': ['DIMENSION(2)'], 'selector': (None, '8'), \
#                'entity_decls': ['kgen_papi_values']}
#            part_append_genknode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)
#
#            attrs = {'type_spec': 'INTEGER', 'entity_decls': ['kgen_retval']}
#            part_append_genknode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)
#
#            part_append_comment(node, DECL_PART, '#else', style='rawtext')

        attrs = {'type_spec': 'INTEGER', 'selector': ('8', None), \
            'entity_decls': ['kgen_start_clock', 'kgen_stop_clock', 'kgen_rate_clock']}
        part_append_genknode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs) 

        attrs = {'type_spec': 'REAL', 'selector': (None, 'kgen_dp'), 'entity_decls': ['gkgen_measure']}
        part_append_genknode(node, DECL_PART, typedecl_statements.Real, attrs=attrs) 

#        if getinfo('is_papi_enabled'):
#            part_append_comment(node, DECL_PART, '#endif', style='rawtext')

        #attrs = {'type_spec': 'INTEGER', 'attrspec': ['PARAMETER'], 'entity_decls': ['kgen_maxiter = %s'%getinfo('repeat_count')]}
        #part_append_genknode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs) 

        prenode = getinfo('blocknode_aftercallsite_main')
        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.BEGIN_PROCESS, \
            prenode, None, self.add_execblock)

    def add_execblock(self, node):

#        if getinfo('is_papi_enabled'):
#            part_append_comment(node, EXEC_PART, '#ifdef KGEN_PAPI', style='rawtext')
#
#            attrs = {'designator': 'PAPIF_start_counters', 'items': ['kgen_papi_events', '1', 'kgen_retval']}
#            part_append_genknode(node, EXEC_PART, statements.Call, attrs=attrs)
#
#            part_append_comment(node, EXEC_PART, '#else', style='rawtext')

        if getinfo('cache_pollution'):

            attrs = {'variable': 'kgen_cache_control', 'sign': '=', 'expr': 'CACHE_CONTROL_LENGTH'}
            part_append_genknode(node, EXEC_PART, statements.Assignment, attrs=attrs)
           
            attrs = {'loopcontrol': 'kgen_intvar = CACHE_CONTROL_LENGTH-1, 1, -1'}
            doobj = part_append_genknode(node, EXEC_PART, block_statements.Do, attrs=attrs)

            attrs = {'variable': 'kgen_cache_control(kgen_intvar)', 'sign': '=', 'expr': 'kgen_cache_control(kgen_intvar+1) - 1 '}
            part_append_genknode(doobj, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'designator': 'SYSTEM_CLOCK', 'items': ['kgen_start_clock', 'kgen_rate_clock']}
        part_append_genknode(node, EXEC_PART, statements.Call, attrs=attrs)
       
        attrs = {'loopcontrol': 'kgen_intvar = 1, KGEN_MAXITER'}
        doiter = part_append_genknode(node, EXEC_PART, block_statements.Do, attrs=attrs)

#        if getinfo('is_papi_enabled'):
#            part_append_comment(node, EXEC_PART, '#endif', style='rawtext')

        #execpart = get_part(node, EXEC_PART)
        #namedpart_create_subpart(doobj, KERNEL_PBLOCK_TIMING, EXEC_PART)

        kernel_stmts = getinfo('callsite_stmts')
        start = kernel_stmts[0].item.span[0]-1
        end = kernel_stmts[-1].item.span[1]
        lines = kernel_stmts[0].top.prep[start:end]
        lines_str = '\n'.join(lines)
        dummy_node = part_append_genknode(doiter, EXEC_PART, statements.Call)
        dummy_node.kgen_stmt = getinfo('dummy_stmt')
        dummy_node.kgen_forced_line = lines_str

        #namedpart_append_comment(node.kgen_kernel_id, KERNEL_PBLOCK_TIMING, 'TEST!!!!')

        #attrs = {'variable': 'kgen_resetinvoke', 'sign': '=', 'expr': '.TRUE.'}
        #namedpart_append_gensnode(node.kgen_kernel_id, KERNEL_PBLOCK_TIMING, statements.Assignment, attrs=attrs)
        #for elem in execpart:
        #    print 'AAA', hasattr(elem, 'kgen_stmt') , self.ispstmt(getinfo('callsite_stmts')[0], elem.kgen_stmt, node.kgen_stmt)    
        #    if hasattr(elem, 'kgen_stmt') and self.ispstmt(getinfo('callsite_stmts')[0], elem.kgen_stmt, node.kgen_stmt):
        #        namedpart_append_node(node.kgen_kernel_id, KERNEL_PBLOCK_TIMING, elem)
        #import pdb; pdb.set_trace()

#        if getinfo('is_papi_enabled'):
#            part_append_comment(node, EXEC_PART, '#ifdef KGEN_PAPI', style='rawtext')
#
#            attrs = {'designator': 'PAPIF_read_counters', 'items': ['kgen_papi_values(1)', '1', 'kgen_retval']}
#            part_append_genknode(node, EXEC_PART, statements.Call, attrs=attrs)
#
#            attrs = {'designator': 'PAPIF_stop_counters', 'items': ['kgen_papi_values(2)', '1', 'kgen_retval']}
#            part_append_genknode(node, EXEC_PART, statements.Call, attrs=attrs)
#
#            attrs = {'variable': 'kgen_measure', 'sign': '=', 'expr': 'kgen_papi_values(1)'}
#            part_append_genknode(node, EXEC_PART, statements.Assignment, attrs=attrs)
#
#            attrs = {'items': ['"%s : KGENPAPIEVENT per call: "'%getinfo('kernel_name'), 'kgen_measure']}
#            part_append_gensnode(node, EXEC_PART, statements.Write, attrs=attrs)
#
#            part_append_comment(node, EXEC_PART, '#else', style='rawtext')

        attrs = {'designator': 'SYSTEM_CLOCK', 'items': ['kgen_stop_clock', 'kgen_rate_clock']}
        part_append_genknode(node, EXEC_PART, statements.Call, attrs=attrs)

        if getinfo('cache_pollution'):
            attrs = {'expr': 'kgen_cache_control(1) == 0'}
            ifcache = part_append_genknode(node, EXEC_PART, block_statements.IfThen, attrs=attrs)

            attrs = {'variable': 'kgen_stop_clock', 'sign': '=', 'expr': 'kgen_stop_clock  + kgen_cache_control(1)'}
            part_append_genknode(ifcache, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'variable': 'kgen_measure', 'sign': '=', 'expr': '1.0D6*(kgen_stop_clock - kgen_start_clock)/DBLE(kgen_rate_clock*KGEN_MAXITER)'}
        part_append_genknode(node, EXEC_PART, statements.Assignment, attrs=attrs)

        if getinfo('add_mpi_frame'):
            part_append_comment(node, EXEC_PART, '#ifdef _MPI', style='rawtext')
            part_append_comment(node, EXEC_PART, 'CALL mpi_allreduce(kgen_measure, gkgen_measure, 1, mpi_real8, mpi_max, mpi_comm_world, kgen_ierr)', style='rawtext')
            attrs = {'variable': 'kgen_measure', 'sign': '=', 'expr': 'gkgen_measure'}
            part_append_genknode(node, EXEC_PART, statements.Assignment, attrs=attrs)
            part_append_comment(node, EXEC_PART, '#endif', style='rawtext')

        attrs = {'expr': 'check_status%rank==0'}
        ifrank = part_append_genknode(node, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'items': ['"%s : Time per call (usec): "'%getinfo('kernel_name'), 'kgen_measure']}
        part_append_gensnode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

#        if getinfo('is_papi_enabled'):
#            part_append_comment(node, EXEC_PART, '#endif', style='rawtext')

